""".. Ignore pydocstyle D400.

================
Celery Connector
================

"""
import logging
import sys

from django.conf import settings

from resolwe.flow.models import Process
from resolwe.flow.tasks import celery_run
from resolwe.utils import BraceMessage as __

from .base import BaseConnector

logger = logging.getLogger(__name__)

# Sphinx directly imports the modules it's documenting, so we need to
# guard from importing celery on installations which are configured to
# not use celery and thus don't have it available.
if "sphinx" not in sys.modules:
    try:
        import celery  # noqa: F401
    except ImportError:
        logger.error(
            "Please install Celery using 'pip install celery'", file=sys.stderr
        )
        sys.exit(1)


class Connector(BaseConnector):
    """Celery-based connector for job execution."""

    def submit(self, data, runtime_dir, argv):
        """Run process.

        For details, see
        :meth:`~resolwe.flow.managers.workload_connectors.base.BaseConnector.submit`.
        """
        queue = "ordinary"
        if data.process.scheduling_class == Process.SCHEDULING_CLASS_INTERACTIVE:
            queue = "hipri"

        logger.debug(
            __(
                "Connector '{}' running for Data with id {} ({}) in celery queue {}, EAGER is {}.",
                self.__class__.__module__,
                data.id,
                repr(argv),
                queue,
                getattr(settings, "CELERY_ALWAYS_EAGER", None),
            )
        )
        celery_run.apply_async((data.id, runtime_dir, argv), queue=queue)

    def cleanup(self, data_id: int):
        """Cleanup."""



# pylint: disable=missing-docstring
import random
import time
from concurrent.futures import ThreadPoolExecutor

from resolwe.storage.connectors.transfer import CircularBuffer
from resolwe.test import TestCase


class CircularBufferTest(TestCase):
    def test_basic_read_write(self):
        stream = CircularBuffer()
        data = b"testing"
        stream.write(data)
        self.assertEqual(stream.read(len(data)), data)
        self.assertEqual(stream.tell(), len(data))

    def test_read_write_circular(self):
        stream = CircularBuffer(buffer_size=11)
        data = b"testing"
        stream.write(data)
        self.assertEqual(stream.read(len(data)), data)
        # Here we have to rotate around
        stream.write(data)
        self.assertEqual(stream.read(len(data)), data)
        self.assertEqual(stream.tell(), 2 * len(data))

    def test_read_large_chunk(self):
        reading_bytes = 10000
        write_data = b"small"
        stream = CircularBuffer(buffer_size=11)

        def task_a():
            return stream.read(reading_bytes)

        def task_b():
            for i in range(int(reading_bytes / len(write_data)) + 2):
                stream.write(write_data)

        with ThreadPoolExecutor(max_workers=2) as executor:
            future_a = executor.submit(task_a)
            executor.submit(task_b)
        read_data = future_a.result()
        self.assertEqual(
            read_data,
            (write_data * (int(reading_bytes / len(write_data)) + 1))[:reading_bytes],
        )

    def test_write_large_chunk(self):
        reading_bytes = 1000
        write_data = b"writinginlargechunksmuchlargerthanmybuffer"
        stream = CircularBuffer(buffer_size=11)

        def task_a():
            data = stream.read(reading_bytes)
            stream.close()
            return data

        def task_b():
            for i in range(int(reading_bytes / len(write_data)) + 2):
                stream.write(write_data)

        with ThreadPoolExecutor(max_workers=2) as executor:
            future_a = executor.submit(task_a)
            executor.submit(task_b)
        read_data = future_a.result()
        self.assertEqual(
            read_data,
            (write_data * (int(reading_bytes / len(write_data)) + 1))[:reading_bytes],
        )

    def test_read_write_largesize_delay(self):
        mb = 1024 * 1024
        reading_bytes = 1024 * mb  # 1GB, needs at least 2G RAM
        read_chunk_max_size = 50 * mb  # 50 MB
        write_data = b"testing"
        stream = CircularBuffer(buffer_size=1024 * 1024 * 100)  # 100 MB

        def task_a():
            read_bytes = 0
            data = bytearray(reading_bytes)
            while read_bytes < reading_bytes:
                time.sleep(random.random() / 10)  # max 0.1s
                to_read = random.randint(1, read_chunk_max_size)
                to_read = min(to_read, reading_bytes - read_bytes)
                chunk = stream.read(to_read)
                data[read_bytes : read_bytes + to_read] = chunk
                read_bytes += to_read
            stream.close()
            return data

        def task_b():
            wrote_bytes = 0
            while wrote_bytes < reading_bytes:
                time.sleep(random.random() / 10)  # max 0.1s
                chunk = write_data * random.randint(1, 1000000)
                stream.write(chunk)
                wrote_bytes += len(chunk)

        with ThreadPoolExecutor(max_workers=2) as executor:
            future_a = executor.submit(task_a)
            executor.submit(task_b)
        read_data = future_a.result()
        self.assertEqual(
            read_data,
            (write_data * (int(reading_bytes / len(write_data)) + 1))[:reading_bytes],
        )



"""Global settings for Flow executor."""
from pathlib import Path
from typing import Any, Dict

from .protocol import ExecutorFiles

DESERIALIZED_FILES: Dict[str, Any] = {
    getattr(ExecutorFiles, f): {} for f in dir(ExecutorFiles) if f == f.upper()
}


def initialize_constants(data_id: int, bootstrap_data: Dict):
    """Update the constants with the bootstrap data."""
    global DATA_ID, EXECUTOR_SETTINGS, SETTINGS, LOCATION_SUBPATH, PROCESS, PROCESS_META
    EXECUTOR_SETTINGS = bootstrap_data.get(ExecutorFiles.EXECUTOR_SETTINGS)
    SETTINGS = bootstrap_data.get(ExecutorFiles.DJANGO_SETTINGS)
    LOCATION_SUBPATH = Path(bootstrap_data.get(ExecutorFiles.LOCATION_SUBPATH, "-1"))
    PROCESS = bootstrap_data.get(ExecutorFiles.PROCESS)
    PROCESS_META = bootstrap_data.get(ExecutorFiles.PROCESS_META)
    DATA_ID = data_id


EXECUTOR_SETTINGS = DESERIALIZED_FILES[ExecutorFiles.EXECUTOR_SETTINGS]
SETTINGS = DESERIALIZED_FILES[ExecutorFiles.DJANGO_SETTINGS]
LOCATION_SUBPATH = DESERIALIZED_FILES[ExecutorFiles.LOCATION_SUBPATH]
PROCESS = DESERIALIZED_FILES[ExecutorFiles.PROCESS]
PROCESS_META = DESERIALIZED_FILES[ExecutorFiles.PROCESS_META]
DATA_ID = -1



"""An execution engine for Python processes."""
import os

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.process.parser import SafeParser

PYTHON_RUNTIME_DIRNAME = "python_runtime"
PYTHON_RUNTIME_ROOT = "/"
PYTHON_RUNTIME_VOLUME = os.path.join(PYTHON_RUNTIME_ROOT, PYTHON_RUNTIME_DIRNAME)


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that outputs bash programs."""

    name = "python"

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        if not path.lower().endswith(".py"):
            return []
        parser = SafeParser(open(path).read())
        processes = parser.parse()
        return [process.to_schema() for process in processes]

    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        return 'PYTHONPATH="{runtime}" python3 -u -m resolwe.process'.format(
            runtime=PYTHON_RUNTIME_VOLUME
        )

    def prepare_volumes(self):
        """Mount additional volumes."""
        return {PYTHON_RUNTIME_DIRNAME: PYTHON_RUNTIME_VOLUME}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-23 09:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0035_process_entity_new"),
    ]

    operations = [
        migrations.AddField(
            model_name="data",
            name="collection",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="data2",
                to="flow.Collection",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="entity2",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="data2",
                to="flow.Entity",
            ),
        ),
        migrations.AddField(
            model_name="entity",
            name="collection",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="entities2",
                to="flow.Collection",
            ),
        ),
    ]



# Generated by Django 2.2.15 on 2020-10-05 14:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0046_purge_data_dependencies"),
    ]

    operations = [
        migrations.AlterField(
            model_name="datadependency",
            name="kind",
            field=models.CharField(
                choices=[
                    ("io", "Input/output dependency"),
                    ("subprocess", "Subprocess"),
                    ("duplicate", "Duplicate"),
                ],
                max_length=16,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-07-28 06:17
from __future__ import unicode_literals

from django.db import migrations

from django.contrib.auth import get_user_model

from resolwe.permissions.utils import assign_contributor_permissions


def set_processes_owner(apps, schema_editor):
    user_model = get_user_model()

    Process = apps.get_model("flow", "Process")
    for process in Process.objects.all():
        # In migrations contributor is not a real user instance, so we
        # have to get it.
        user = user_model.objects.get(pk=process.contributor.pk)
        assign_contributor_permissions(process, user)


def set_descriptor_owner(apps, schema_editor):
    user_model = get_user_model()

    DescriptorSchema = apps.get_model("flow", "DescriptorSchema")
    for descriptor in DescriptorSchema.objects.all():
        # In migrations contributor is not a real user instance, so we
        # have to get it.
        user = user_model.objects.get(pk=descriptor.contributor.pk)
        assign_contributor_permissions(descriptor, user)


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0001_squashed_0030_change_slug_field"),
    ]

    operations = [
        migrations.RunPython(set_processes_owner),
        migrations.RunPython(set_descriptor_owner),
    ]



"""Resolwe entity model."""
from django.contrib.postgres.fields import CICharField
from django.contrib.postgres.indexes import GinIndex
from django.db import models, transaction

from resolwe.permissions.utils import copy_permissions

from .base import BaseModel, BaseQuerySet
from .collection import BaseCollection
from .utils import bulk_duplicate


class EntityQuerySet(BaseQuerySet):
    """Query set for ``Entity`` objects."""

    @transaction.atomic
    def duplicate(self, contributor, inherit_collection=False):
        """Duplicate (make a copy) ``Entity`` objects.

        :param contributor: Duplication user
        :param inherit_collection: If ``True`` then duplicated
            entities will be added to collection the original entity
            is part of. Duplicated entities' data objects will also be
            added to the collection, but only those which are in the
            collection
        :return: A list of duplicated entities
        """
        return bulk_duplicate(
            entities=self,
            contributor=contributor,
            inherit_collection=inherit_collection,
        )

    @transaction.atomic
    def move_to_collection(self, source_collection, destination_collection):
        """Move entities from source to destination collection."""
        for entity in self:
            entity.move_to_collection(source_collection, destination_collection)


class Entity(BaseCollection):
    """Postgres model for storing entities."""

    class Meta(BaseCollection.Meta):
        """Entity Meta options."""

        permissions = (
            ("view_entity", "Can view entity"),
            ("edit_entity", "Can edit entity"),
            ("share_entity", "Can share entity"),
            ("owner_entity", "Is owner of the entity"),
        )

        indexes = [
            models.Index(name="idx_entity_name", fields=["name"]),
            GinIndex(
                name="idx_entity_name_trgm", fields=["name"], opclasses=["gin_trgm_ops"]
            ),
            models.Index(name="idx_entity_slug", fields=["slug"]),
            GinIndex(name="idx_entity_tags", fields=["tags"]),
            GinIndex(name="idx_entity_search", fields=["search"]),
        ]

    #: manager
    objects = EntityQuerySet.as_manager()

    #: collection to which entity belongs
    collection = models.ForeignKey(
        "Collection", blank=True, null=True, on_delete=models.CASCADE
    )

    #: entity type
    type = models.CharField(max_length=100, db_index=True, null=True, blank=True)

    #: duplication date and time
    duplicated = models.DateTimeField(blank=True, null=True)

    def is_duplicate(self):
        """Return True if entity is a duplicate."""
        return bool(self.duplicated)

    def duplicate(self, contributor, inherit_collection=False):
        """Duplicate (make a copy)."""
        return bulk_duplicate(
            entities=self._meta.model.objects.filter(pk=self.pk),
            contributor=contributor,
            inherit_collection=inherit_collection,
        )[0]

    @transaction.atomic
    def move_to_collection(self, source_collection, destination_collection):
        """Move entity to destination collection."""
        if source_collection == destination_collection:
            return

        self.collection = destination_collection
        if destination_collection:
            self.tags = destination_collection.tags
            copy_permissions(destination_collection, self)
        self.save()

        for datum in self.data.all():
            datum.collection = destination_collection
            if destination_collection:
                datum.tags = destination_collection.tags
                copy_permissions(destination_collection, datum)
            datum.save()


class RelationType(models.Model):
    """Model for storing relation types."""

    #: relation type name
    name = models.CharField(max_length=100, unique=True)

    #: indicates if order of entities in relation is important or not
    ordered = models.BooleanField(default=False)

    def __str__(self):
        """Format the object representation."""
        return self.name


class Relation(BaseModel):
    """Relations between entities.

    The Relation model defines the associations and dependencies between
    entities in a given collection:

    .. code-block:: json

        {
            "collection": "<collection_id>",
            "type": "comparison",
            "category": "case-control study",
            "entities": [
                {"enetity": "<entity1_id>", "label": "control"},
                {"enetity": "<entity2_id>", "label": "case"},
                {"enetity": "<entity3_id>", "label": "case"}
            ]
        }

    Relation ``type`` defines a specific set of associations among
    entities. It can be something like ``group``, ``comparison`` or
    ``series``. The relation ``type`` is an instance of
    :class:`.RelationType` and should be defined in any Django app that
    uses relations (e.g., as a fixture). Multiple relations of the same
    type are allowed on the collection.

    Relation ``category`` defines a specific use case. The relation
    category must be unique in a collection, so that users can
    distinguish between different relations. In the example above, we
    could add another ``comparison`` relation of ``category``, say
    ``Case-case study`` to compare ``<entity2>`` with ``<entity3>``.

    Relation is linked to :class:`resolwe.flow.models.Collection` to
    enable defining different relations structures in different
    collections. This also greatly speed up retrieving of relations, as
    they are envisioned to be mainly used on a collection level.

    ``unit`` defines units used in partitions where it is applicable,
    e.g. in relations of type ``series``.

    """

    UNIT_SECOND = "s"
    UNIT_MINUTE = "min"
    UNIT_HOUR = "hr"
    UNIT_DAY = "d"
    UNIT_WEEK = "wk"
    UNIT_CHOICES = (
        (UNIT_SECOND, "Second"),
        (UNIT_MINUTE, "Minute"),
        (UNIT_HOUR, "Hour"),
        (UNIT_DAY, "Day"),
        (UNIT_WEEK, "Week"),
    )

    #: type of the relation
    type = models.ForeignKey("RelationType", on_delete=models.PROTECT)

    #: collection to which relation belongs
    collection = models.ForeignKey("Collection", on_delete=models.CASCADE)

    #: partitions of entities in the relation
    entities = models.ManyToManyField(Entity, through="RelationPartition")

    #: category of the relation
    category = CICharField(max_length=100)

    #: unit used in the partitions' positions (where applicable, e.g. for serieses)
    unit = models.CharField(max_length=3, choices=UNIT_CHOICES, null=True, blank=True)

    class Meta(BaseModel.Meta):
        """Relation Meta options."""

        unique_together = ("collection", "category")

        permissions = (
            ("view_relation", "Can view relation"),
            ("edit_relation", "Can edit relation"),
            ("share_relation", "Can share relation"),
            ("owner_relation", "Is owner of the relation"),
        )


class RelationPartition(models.Model):
    """Intermediate model to define Relations between Entities."""

    #: entity in the relation
    entity = models.ForeignKey(Entity, on_delete=models.CASCADE)

    #: relation to which entity belongs
    relation = models.ForeignKey(Relation, on_delete=models.CASCADE)

    #: number used for ordering entities in the relation
    position = models.PositiveSmallIntegerField(null=True, blank=True, db_index=True)

    #: Human-readable label of the position
    label = models.CharField(max_length=30, null=True, blank=True, db_index=True)

    class Meta:
        """RelationPartition Meta options."""

        unique_together = ("entity", "relation")



""".. Ignore pydocstyle D400.

===================
Resolwe Test Runner
===================

"""
import asyncio
import contextlib
import errno
import inspect
import logging
import os
import re
import shutil
import subprocess
import sys
from importlib import import_module
from pathlib import Path
from unittest.mock import patch

import yaml
import zmq
import zmq.asyncio
from channels.db import database_sync_to_async

from django.conf import settings
from django.core.management.base import CommandError
from django.test import override_settings
from django.test.runner import DiscoverRunner, ParallelTestSuite, RemoteTestRunner
from django.utils.crypto import get_random_string

# Make sure we already have the patched FLOW_* available here; otherwise
# resolwe.test.testcases.TransactionTestCase will override them with the module above,
# negating anything we do here with Django's override_settings.
import resolwe.test.testcases.setting_overrides as resolwe_settings
from resolwe.flow.finders import get_finders
from resolwe.flow.managers import listener, manager, state
from resolwe.storage.connectors import connectors
from resolwe.test.utils import generate_process_tag

logger = logging.getLogger(__name__)

SPAWN_PROCESS_REGEX = re.compile(
    r'run\s+\{.*?["\']process["\']\s*:\s*["\'](.+?)["\'].*?\}'
)

TESTING_CONTEXT = {
    "is_testing": False,
}


class TestingContext:
    """Context manager which maintains current testing status."""

    def __enter__(self):
        """Enter testing context."""
        TESTING_CONTEXT["is_testing"] = True

    def __exit__(self, *args, **kwargs):
        """Exit testing context."""
        TESTING_CONTEXT["is_testing"] = False

        # Propagate exceptions.
        return False


def _manager_setup():
    """Execute setup operations common to serial and parallel testing.

    This mostly means state cleanup, such as resetting database
    connections and clearing the shared state.
    """
    if TESTING_CONTEXT.get("manager_reset", False):
        return
    TESTING_CONTEXT["manager_reset"] = True
    state.update_constants()
    manager.drain_messages()


def _sequence_paths(paths):
    """Extend the last components of the given paths with a number.

    The method finds the lowest number such that all given paths, when
    extended by it, are unique and can be created. The paths are then
    also created.

    :param paths: The list of paths to be extended and created.
    :return: The list of created paths.
    """
    seq = 0
    while True:
        # Note for parallel execution: infinite zigzagging ladders are
        # not possible, because the directories are always created in
        # the same order. The problem would be if process A succeeded
        # in creating data/test_1, but process B would beat it to
        # upload/test_1 (so that both would roll back and continue
        # with _2, etc.). For B to succeed in creating upload/test_1,
        # it must have already succeeded in creating data/test_1,
        # meaning A could not possibly have succeeded with data/test_1.
        seq += 1
        created = []
        hashseed = os.environ.get("PYTHONHASHSEED", "")

        for base_path in paths:
            path = os.path.join(base_path, "test_{}_{}".format(hashseed, seq))
            try:
                os.makedirs(path)
                created.append(path)
            except OSError as e:
                # Retry on file exists error.
                if e.errno == errno.EEXIST:
                    break
                raise

        if len(created) == len(paths):
            return created

        # If they're not equal, we failed and need to roll back;
        # errors are entirely irrelevant here, removal is purely
        # best effort.
        for path in created:
            try:
                os.rmdir(path)
            except Exception:
                pass


def _create_test_dirs():
    """Create all the testing directories."""
    if "test_paths" in TESTING_CONTEXT:
        return TESTING_CONTEXT["test_paths"]
    items = [
        resolwe_settings.STORAGE_CONNECTORS[connector_name]
        for connector_name in resolwe_settings.STORAGE_CONNECTORS
        if connectors[connector_name].mountable
    ]
    items += [
        volume
        for volume in resolwe_settings.FLOW_VOLUMES.values()
        if "read_only" not in volume["config"]
        and volume["type"] != "temporary_directory"
    ]
    paths = _sequence_paths([item["config"]["path"] for item in items])
    for item, path in zip(items, paths):
        item["config"]["path"] = path
    TESTING_CONTEXT["test_paths"] = paths
    return paths


def _prepare_settings():
    """Prepare and apply settings/port overrides needed for testing.

    Override necessary settings and binds to a free port that will be used in
    listener.

    :returns: tuple (overrides, port).
    """
    # Override container name prefix setting.
    mountable_data_connectors = [
        connector for connector in connectors.for_storage("data") if connector.mountable
    ]
    resolwe_settings.FLOW_EXECUTOR_SETTINGS[
        "CONTAINER_NAME_PREFIX"
    ] = "{}_{}_{}".format(
        getattr(settings, "FLOW_EXECUTOR", {}).get("CONTAINER_NAME_PREFIX", "resolwe"),
        # NOTE: This is necessary to avoid container name clashes when tests are run from
        # different Resolwe code bases on the same system (e.g. on a CI server).
        get_random_string(length=6),
        os.path.basename(mountable_data_connectors[0].path),
    )

    hosts = list(
        settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {})
        .get("hosts", {"local": "127.0.0.1"})
        .values()
    )
    protocol = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
        "protocol", "tcp"
    )
    min_port = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
        "min_port", 50000
    )
    max_port = settings.FLOW_EXECUTOR.get("LISTENER_CONNECTION", {}).get(
        "max_port", 60000
    )

    zmq_context: zmq.asyncio.Context = zmq.asyncio.Context.instance()
    zmq_socket: zmq.asyncio.Socket = zmq_context.socket(zmq.ROUTER)
    zmq_socket.setsockopt(zmq.ROUTER_HANDOVER, 1)

    host = hosts[0]
    port = zmq_socket.bind_to_random_port(
        f"{protocol}://{host}", min_port=min_port, max_port=max_port
    )
    for host in hosts[1:]:
        zmq_socket.bind(f"{protocol}://{host}:{port}")

    # Set the port in the settings.
    resolwe_settings.FLOW_EXECUTOR_SETTINGS["LISTENER_CONNECTION"]["port"] = port

    overrides = override_settings(
        CELERY_ALWAYS_EAGER=True,
        FLOW_EXECUTOR=resolwe_settings.FLOW_EXECUTOR_SETTINGS,
        FLOW_MANAGER=resolwe_settings.FLOW_MANAGER_SETTINGS,
    )
    return (overrides, zmq_socket)


def _custom_worker_init(django_init_worker):
    """Wrap the original worker init to also start the manager."""

    def _init_worker(*args, **kwargs):
        """Initialize a :class:`multiprocessing.Pool` worker.

        Call the Django's ``ParallelTestSuite.init_worker`` and then
        also start the manager infrastructure.
        """
        result = django_init_worker(*args, **kwargs)

        # Further patch channel names and the like with our current pid,
        # so that parallel managers and executors don't clash on the
        # same channels and directories.
        resolwe_settings.FLOW_MANAGER_SETTINGS[
            "REDIS_PREFIX"
        ] += "-parallel-pid{}".format(os.getpid())

        return result

    return _init_worker


def _run_in_event_loop(coro, *args, **kwargs):
    """Run a coroutine in a runloop call.

    This is needed as the top level call into Resolwe Manager-using
    tests. An event loop is started so that it can be used within the
    call tree.

    :param coro: The coroutine to run with an underlying event loop. All
        other arguments given to this function are forwarded to it.
    """
    asyncio.get_event_loop().close()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    task = asyncio.ensure_future(coro(*args, **kwargs), loop=loop)
    loop.run_until_complete(task)
    loop.close()
    return task.result()


async def _run_on_infrastructure(meth, *args, **kwargs):
    """Start the Manager infrastructure and call the given callable.

    The method given is run through a serializing wrapper, so that
    Django database accesses are correct.

    :param meth: The callable to run on the infrastructure. All other
        arguments are forwarded to it.
    """
    with TestingContext():
        _create_test_dirs()
        overrides, zmq_socket = _prepare_settings()
        with overrides:
            with patch.multiple(
                "resolwe.storage.settings",
                STORAGE_CONNECTORS=resolwe_settings.STORAGE_CONNECTORS,
                FLOW_VOLUMES=resolwe_settings.FLOW_VOLUMES,
            ):
                connectors.recreate_connectors()
                await database_sync_to_async(_manager_setup)()
                hosts = settings.FLOW_EXECUTOR["LISTENER_CONNECTION"]["hosts"]
                port = settings.FLOW_EXECUTOR["LISTENER_CONNECTION"]["port"]
                protocol = settings.FLOW_EXECUTOR["LISTENER_CONNECTION"]["protocol"]
                listener.hosts = hosts
                listener.port = port
                listener.protocol = protocol
                listener.zmq_socket = zmq_socket
                async with listener:
                    try:
                        with override_settings(FLOW_MANAGER_SYNC_AUTO_CALLS=True):
                            # Run the test in the new thread instead on the
                            # main thread (default). If test is started on the
                            # main thread then database_sync_to_async calls
                            # will wait indefinitely (they share the same
                            # thread) for the test to finish creating the
                            # infinite loop.
                            result = await database_sync_to_async(
                                meth, thread_sensitive=False
                            )(*args, **kwargs)
                        return result
                    except Exception:
                        logger.exception("Exception while running test")
                    finally:
                        logger.debug("test_runner: Terminating listener")


def _run_manager(meth, *args, **kwargs):
    """Start the Manager properly and nest the given callable in it.

    :param meth: The callable to start the environment for; all other
        arguments are forwarded to it.
    """
    return _run_in_event_loop(_run_on_infrastructure, meth, *args, **kwargs)


class CustomRemoteRunner(RemoteTestRunner):
    """Standard Django remote runner with a custom run method."""

    def run(self, *args, **kwargs):
        """Run the superclass method with an underlying event loop."""
        # NOTE: An instance of this class is constructed for _each_ batch
        # of tests, not just once per pool worker, so some care needs to
        # be taken to avoid incremental corruption, such as in
        # _create_test_dirs().
        return _run_manager(super().run, *args, **kwargs)


class CustomParallelTestSuite(ParallelTestSuite):
    """Standard parallel suite with a custom worker initializer."""

    init_worker = _custom_worker_init(ParallelTestSuite.init_worker)
    runner_class = CustomRemoteRunner


class ResolweRunner(DiscoverRunner):
    """Resolwe test runner."""

    parallel_test_suite = CustomParallelTestSuite

    def __init__(self, *args, **kwargs):
        """Initialize test runner."""
        self.only_changes_to = kwargs.pop("only_changes_to", None)
        self.changes_file_types = kwargs.pop("changes_file_types", None)
        self.keep_data = kwargs.pop("keep_data", False)

        super().__init__(*args, **kwargs)

    @classmethod
    def add_arguments(cls, parser):
        """Add command-line arguments.

        :param parser: Argument parser instance
        """
        super().add_arguments(parser)

        parser.add_argument(
            "--only-changes-to",
            dest="only_changes_to",
            help="Only test changes against given Git commit reference",
        )

        parser.add_argument(
            "--changes-file-types",
            dest="changes_file_types",
            help="File which describes what kind of changes are available",
        )

        parser.add_argument(
            "--keep-data",
            dest="keep_data",
            action="store_true",
            help="Prevent test cases from cleaning up after execution",
        )

    def build_suite(self, *args, **kwargs):
        """Build test suite."""
        suite = super().build_suite(*args, **kwargs)
        # Build suite first constructs the parallel suite and then may reduce self.parallel,
        # while keeping suite.processes unchanged. We need to propagate the change here to
        # avoid spawning more processes than there are databases.
        suite.processes = self.parallel

        # Augment all test cases with manager state validation logic.
        def validate_manager_state(case, teardown):
            """Decorate test case with manager state validation."""

            def wrapper(*args, **kwargs):
                """Validate manager state on teardown."""
                teardown(*args, **kwargs)

            return wrapper

        if isinstance(suite, self.parallel_test_suite):
            # NOTE: validate_manager_state function cannot be pickled, so it
            #       cannot be used in parallel tests.
            pass
        else:
            for case in suite:
                case.tearDown = validate_manager_state(case, case.tearDown)

        return suite

    def run_suite(self, suite, **kwargs):
        """Run the test suite with manager workers in the background."""
        # Due to the way the app modules are imported, there's no way to
        # statically override settings with TEST overrides before e.g.
        # resolwe.flow.managers.manager is loaded somewhere in the code
        # (the most problematic files are signals.py and
        # resolwe.test.testcases.process); the only realistic mechanism
        # is to override later and call some sort of commit method in
        # the manager.
        keep_data_override = override_settings(FLOW_MANAGER_KEEP_DATA=self.keep_data)
        keep_data_override.__enter__()

        # Prepare the runtime environment. It should contain a copy of the
        # executors runtime and connectors runtime inside it. The runtime
        # can be read-only and can be prepared only once per the entire test
        # suite.
        import resolwe.flow.executors as executor_package
        import resolwe.storage.connectors as connectors_package

        source_file = inspect.getsourcefile(executor_package)
        assert (
            source_file is not None
        ), "Unable to determine the 'resolwe.flow.executors' source directory."
        exec_dir = Path(source_file).parent
        dest_dir = (
            Path(resolwe_settings.FLOW_VOLUMES["runtime"]["config"]["path"])
            / "executors"
        )
        shutil.rmtree(dest_dir, ignore_errors=True)
        shutil.copytree(exec_dir, dest_dir)

        source_file = inspect.getsourcefile(connectors_package)
        assert (
            source_file is not None
        ), "Unable to determine the 'resolwe.storage.connectors' source directory."
        exec_dir = Path(source_file).parent
        dest_dir = dest_dir / "connectors"
        shutil.copytree(exec_dir, dest_dir)

        for runtime_class_name in settings.FLOW_PROCESSES_RUNTIMES:
            module_name, class_name = runtime_class_name.rsplit(".", 1)
            runtime_module = import_module(module_name)
            source_file = Path(inspect.getsourcefile(runtime_module))
            source_dir = source_file.parent
            dest_dir = Path(resolwe_settings.FLOW_VOLUMES["runtime"]["config"]["path"])
            dest_dir = dest_dir.joinpath("python_runtime", *module_name.split(".")[:-1])
            shutil.rmtree(dest_dir, ignore_errors=True)
            shutil.copytree(source_dir, dest_dir)

        if self.parallel > 1:
            return super().run_suite(suite, **kwargs)

        return _run_manager(super().run_suite, suite, **kwargs)

    def run_tests(self, test_labels, **kwargs):
        """Run tests.

        :param test_labels: Labels of tests to run
        """
        if self.only_changes_to:
            # Check if there are changed files. We need to be able to switch between branches so we
            # can correctly detect changes.
            repo_status = self._git("status", "--porcelain", "--untracked-files=no")
            if repo_status:
                print(
                    "ERROR: Git repository is not clean. Running tests with --only-changes-to",
                    file=sys.stderr,
                )
                print(
                    "       requires that the current Git repository is clean.",
                    file=sys.stderr,
                )
                return False

            print(
                "Detecting changed files between {} and HEAD.".format(
                    self.only_changes_to
                )
            )
            changed_files = self._git(
                "diff", "--name-only", self.only_changes_to, "HEAD"
            )
            changed_files = changed_files.strip().split("\n")
            changed_files = [file for file in changed_files if file.strip()]

            top_level_path = self._git("rev-parse", "--show-toplevel")
            top_level_path = top_level_path.strip()

            # Process changed files to discover what they are.
            changed_files, tags, tests, full_suite = self.process_changed_files(
                changed_files, top_level_path
            )
            print("Changed files:")
            for filename, file_type in changed_files:
                print("  {} ({})".format(filename, file_type))

            if not changed_files:
                print("  none")
                print(
                    "No files have been changed, assuming target is HEAD, running full suite."
                )
            elif full_suite:
                print(
                    "Non-test code or unknown files have been modified, running full test suite."
                )
            else:
                # Run specific tests/tags.
                print("Running with following partial tags: {}".format(", ".join(tags)))
                print(
                    "Running with following partial tests: {}".format(", ".join(tests))
                )

                failed_tests = 0

                # First run with specific tags. Since run_tests may modify self.parallel, we need to store
                # it here and restore it later if we also run with specific test labels.
                parallel = self.parallel
                if tags:
                    self.tags = tags
                    failed_tests += super().run_tests(test_labels, **kwargs)

                # Then run with specific test labels.
                if tests:
                    self.parallel = parallel
                    self.tags = set()
                    failed_tests += super().run_tests(tests, **kwargs)

                return failed_tests

        return super().run_tests(test_labels, **kwargs)

    def _git(self, *args):
        """Helper to run Git command."""
        try:
            return subprocess.check_output(["git"] + list(args)).decode("utf8").strip()
        except subprocess.CalledProcessError:
            raise CommandError("Git command failed.")

    @contextlib.contextmanager
    def git_branch(self, branch):
        """Temporarily switch to a different Git branch."""
        current_branch = self._git("rev-parse", "--abbrev-ref", "HEAD")
        if current_branch == "HEAD":
            # Detached HEAD state, we need to get the actual commit.
            current_branch = self._git("rev-parse", "HEAD")

        if current_branch != branch:
            self._git("checkout", branch)

        try:
            yield
        finally:
            if current_branch != branch:
                self._git("checkout", current_branch)

    def process_changed_files(self, changed_files, top_level_path):
        """Process changed files based on specified patterns.

        :param changed_files: A list of changed file pats, relative to top-level path
        :param top_level_path: Absolute path to top-level project directory
        :return: Tuple (changed_files, tags, tests, full_suite)
        """
        result = []
        processes = []
        tests = []
        full_suite = False
        types = []

        if self.changes_file_types:
            # Parse file types metadata.
            try:
                with open(self.changes_file_types, "r") as definition_file:
                    types = yaml.load(definition_file, Loader=yaml.FullLoader)
            except (OSError, ValueError):
                raise CommandError("Failed loading or parsing file types metadata.")
        else:
            print(
                "WARNING: Treating all files as unknown because --changes-file-types option not specified.",
                file=sys.stderr,
            )

        for filename in changed_files:
            # Match file type.
            file_type = "unknown"
            file_type_name = "unknown"

            for definition in types:
                if re.search(definition["match"], filename):
                    file_type = definition["type"]
                    file_type_name = definition.get("name", file_type)
                    break

            result.append((filename, file_type_name))
            if file_type in ("unknown", "force_run"):
                full_suite = True
            elif file_type == "ignore":
                # Ignore
                pass
            elif file_type == "process":
                # Resolve process tag.
                processes.append(os.path.join(top_level_path, filename))
            elif file_type == "test":
                # Generate test name.
                tests.append(re.sub(r"\.py$", "", filename).replace(os.path.sep, "."))
            else:
                raise CommandError("Unsupported file type: {}".format(file_type))

        # Resolve tags.
        tags = self.resolve_process_tags(processes)

        return result, tags, tests, full_suite

    def find_schemas(self, schema_path):
        """Find process schemas.

        :param schema_path: Path where to look for process schemas
        :return: Found schemas
        """
        schema_matches = []

        for root, _, files in os.walk(schema_path):
            for schema_file in [os.path.join(root, fn) for fn in files]:

                if not schema_file.lower().endswith((".yml", ".yaml")):
                    continue

                with open(schema_file) as fn:
                    schemas = yaml.load(fn, Loader=yaml.FullLoader)

                if not schemas:
                    print(
                        "WARNING: Could not read YAML file {}".format(schema_file),
                        file=sys.stderr,
                    )
                    continue

                for schema in schemas:
                    schema_matches.append(schema)

        return schema_matches

    def find_dependencies(self, schemas):
        """Compute process dependencies.

        :param schemas: A list of all discovered process schemas
        :return: Process dependency dictionary
        """
        dependencies = {}

        for schema in schemas:
            slug = schema["slug"]
            run = schema.get("run", {})
            program = run.get("program", None)
            language = run.get("language", None)

            if language == "workflow":
                for step in program:
                    dependencies.setdefault(step["run"], set()).add(slug)
            elif language == "bash":
                # Process re-spawn instructions to discover dependencies.
                matches = SPAWN_PROCESS_REGEX.findall(program)
                if matches:
                    for match in matches:
                        dependencies.setdefault(match, set()).add(slug)

        return dependencies

    def resolve_process_tags(self, files):
        """Resolve process tags.

        :param files: List of changed process files
        :return: Test tags that need to be run
        """
        processes_paths = []
        for finder in get_finders():
            processes_paths.extend(finder.find_processes())

        process_schemas = []
        for proc_path in processes_paths:
            process_schemas.extend(self.find_schemas(proc_path))

        # Switch to source branch and get all the schemas from there as well, since some schemas
        # might have been removed.
        with self.git_branch(self.only_changes_to):
            for proc_path in processes_paths:
                process_schemas.extend(self.find_schemas(proc_path))

        dependencies = self.find_dependencies(process_schemas)
        processes = set()

        def load_process_slugs(filename):
            """Add all process slugs from specified file."""
            with open(filename, "r") as process_file:
                data = yaml.load(process_file, Loader=yaml.FullLoader)

                for process in data:
                    # Add all process slugs.
                    processes.add(process["slug"])

        for filename in files:
            try:
                load_process_slugs(filename)
            except FileNotFoundError:
                # File was removed, so we will handle it below when we check the original branch.
                pass

        # Switch to source branch and check modified files there as well.
        with self.git_branch(self.only_changes_to):
            for filename in files:
                try:
                    load_process_slugs(filename)
                except FileNotFoundError:
                    # File was added, so it has already been handled.
                    pass

        # Add all dependencies.
        dep_processes = set()
        while processes:
            process = processes.pop()
            if process in dep_processes:
                continue

            dep_processes.add(process)
            processes.update(dependencies.get(process, set()))

        tags = set()
        for process in dep_processes:
            tags.add(generate_process_tag(process))

        return tags


def is_testing():
    """Return current testing status."""
    return TESTING_CONTEXT["is_testing"]



# pylint: disable=missing-docstring
from concurrent.futures import ThreadPoolExecutor
from threading import Event
from unittest.mock import MagicMock, call, patch

from django.db import connection, transaction

from resolwe.storage.cleanup import Cleaner
from resolwe.storage.models import FileStorage, StorageLocation
from resolwe.test import TransactionTestCase


class CleanupTest(TransactionTestCase):
    def setUp(self):
        self.file_storage1: FileStorage = FileStorage.objects.create()
        self.file_storage2: FileStorage = FileStorage.objects.create()
        self.cleaner = Cleaner()
        super().setUp()

    def test_process(self):
        process_filestorage_mock = MagicMock()
        with patch(
            "resolwe.storage.cleanup.Cleaner._process_file_storage",
            process_filestorage_mock,
        ):
            self.cleaner.process()
        self.assertEqual(process_filestorage_mock.call_count, 2)
        self.assertIn(call(self.file_storage1), process_filestorage_mock.call_args_list)
        self.assertIn(call(self.file_storage2), process_filestorage_mock.call_args_list)

    def test_skip_locked(self):
        rows_locked = Event()
        cleanup_finished = Event()

        def task_a(lock_ids=[]):
            with transaction.atomic():
                list(FileStorage.objects.select_for_update().filter(id__in=lock_ids))
                rows_locked.set()
                cleanup_finished.wait()
            connection.close()

        def task_b():
            rows_locked.wait()
            self.cleaner.process()
            cleanup_finished.set()
            connection.close()

        process_filestorage_mock = MagicMock()
        with patch(
            "resolwe.storage.cleanup.Cleaner._process_file_storage",
            process_filestorage_mock,
        ):
            with ThreadPoolExecutor() as executor:
                executor.submit(task_a, [self.file_storage1.id, self.file_storage2.id])
                executor.submit(task_b)
        process_filestorage_mock.assert_not_called()

        rows_locked.clear()
        cleanup_finished.clear()
        process_filestorage_mock = MagicMock()
        with patch(
            "resolwe.storage.cleanup.Cleaner._process_file_storage",
            process_filestorage_mock,
        ):
            with ThreadPoolExecutor() as executor:
                executor.submit(task_a, [self.file_storage1.id])
                executor.submit(task_b)
        process_filestorage_mock.assert_called_once_with(self.file_storage2)

    def test_process_file_storage(self):
        delete_mock = MagicMock()
        filter_mock = MagicMock(return_value=[])
        locations_mock = MagicMock(filter=filter_mock, count=lambda: 0)
        file_storage_mock = MagicMock(
            storage_locations=locations_mock, delete=delete_mock
        )
        self.cleaner._process_file_storage(file_storage_mock)
        filter_mock.assert_called_once_with(status=StorageLocation.STATUS_DELETING)
        delete_mock.assert_called_once_with()

        delete_mock = MagicMock()
        filter_mock = MagicMock(return_value=[])
        locations_mock = MagicMock(filter=filter_mock, count=lambda: 1)
        file_storage_mock = MagicMock(
            storage_locations=locations_mock, delete=delete_mock
        )
        self.cleaner._process_file_storage(file_storage_mock)
        filter_mock.assert_called_once_with(status=StorageLocation.STATUS_DELETING)
        delete_mock.assert_not_called()

        delete_mock = MagicMock()
        filter_mock = MagicMock(return_value=["My location"])
        locations_mock = MagicMock(filter=filter_mock, count=lambda: 1)
        file_storage_mock = MagicMock(
            storage_locations=locations_mock, delete=delete_mock
        )
        cleanup_mock = MagicMock()
        with patch(
            "resolwe.storage.cleanup.Cleaner._cleanup",
            cleanup_mock,
        ):
            self.cleaner._process_file_storage(file_storage_mock)
        cleanup_mock.assert_called_once_with("My location")
        filter_mock.assert_called_once_with(status=StorageLocation.STATUS_DELETING)
        delete_mock.assert_not_called()

    def test_delete(self):
        connector_mock = MagicMock()
        delete_mock = MagicMock()
        storage_location = MagicMock(connector=connector_mock, delete=delete_mock)
        self.cleaner._cleanup(storage_location)
        delete_mock.assert_called_once_with()

        connector_mock = MagicMock()
        delete_mock = MagicMock(side_effect=Exception)
        storage_location = MagicMock(connector=connector_mock, delete=delete_mock)
        self.cleaner._cleanup(storage_location)
        delete_mock.assert_called_once_with()



""".. Ignore pydocstyle D400.

=========================
Prepare runtime directory
=========================

"""
import inspect
import shutil
from importlib import import_module
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from resolwe.storage import settings as storage_settings


class Command(BaseCommand):
    """Prepare runtime directory."""

    help = "Prepare runtime directory."

    def __init__(self, *args, **kwargs):
        """Initialize command."""
        super().__init__(*args, **kwargs)

        if not "runtime" in storage_settings.FLOW_VOLUMES:
            raise CommandError("Volume 'runtime' must be defined in storage settings.")

        self.runtime_dir = Path(
            storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"]
        )

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "-c",
            "--clear",
            action="store_true",
            dest="clear",
            help="Clear existing files before copying new files.",
        )

        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Do NOT prompt the user for input of any kind.",
        )

    def get_confirmation(self):
        """Get user confirmation to proceed."""
        if self.clear:
            action = "This will DELETE ALL FILES in this location!"
        else:
            action = "This will overwrite existing files!"

        message = (
            "\n"
            "You have requested to prepare runtime directory at the destination\n"
            "location as specified in your settings\n"
            "\n"
            "    {destination}\n"
            "\n"
            "{action}\n"
            "Are you sure you want to do this?\n"
            "\n"
            "Type 'yes' to continue, or 'no' to cancel: ".format(
                destination=self.runtime_dir,
                action=action,
            )
        )

        if input("".join(message)) != "yes":
            raise CommandError("Preparing runtime directory cancelled.")

    def clear_runtime(self):
        """Delete contents of the runtime directory."""
        self.stdout.write("Deleting contents of '{}'.".format(self.runtime_dir))
        for entry in self.runtime_dir.glob("*"):
            if entry.is_file() or entry.is_symlink():
                entry.unlink()
            elif entry.is_dir():
                shutil.rmtree(entry)

    def copy_dir(self, source: Path, destination: Path):
        """Copy dir from source to destination.

        The shutil copytree method could be used but it will not overwrite
        directories (option added in Python 3.8, while we support Python 3.6).

        This method iterates through all entries in source directory and copies
        them over to the destination directory (overwriting them if they
        already exist).
        """
        for source_entry in source.rglob("*"):
            relative_source_path = source_entry.relative_to(source)
            destination_entry = destination / relative_source_path
            if source_entry.is_dir():
                destination_entry.mkdir(exist_ok=True, parents=True)
            elif source_entry.is_file() or source_entry.is_symlink():
                destination_entry.parent.mkdir(exist_ok=True, parents=True)
                shutil.copy(source_entry, destination_entry)

    def prepare_runtime(self):
        """Prepare runtime directory."""
        import resolwe.flow.executors as executor_package
        import resolwe.storage.connectors as connectors_package

        source_file = inspect.getsourcefile(executor_package)
        assert (
            source_file is not None
        ), "Unable to determine the 'resolwe.flow.executors' source directory."
        exec_dir = Path(source_file).parent
        dest_dir = self.runtime_dir / "executors"
        self.copy_dir(exec_dir, dest_dir)

        source_file = inspect.getsourcefile(connectors_package)
        assert (
            source_file is not None
        ), "Unable to determine the 'resolwe.storage.connectors' source directory."
        exec_dir = Path(source_file).parent
        dest_dir = dest_dir / "connectors"
        self.copy_dir(exec_dir, dest_dir)

        for runtime_class_name in settings.FLOW_PROCESSES_RUNTIMES:
            module_name, class_name = runtime_class_name.rsplit(".", 1)
            runtime_module = import_module(module_name)
            source_file = Path(inspect.getsourcefile(runtime_module))
            source_dir = source_file.parent

            dest_dir = Path(storage_settings.FLOW_VOLUMES["runtime"]["config"]["path"])
            dest_dir = dest_dir.joinpath("python_runtime", *module_name.split(".")[:-1])
            self.copy_dir(source_dir, dest_dir)

    def set_options(self, **options):
        """Set instance variables based on an options dict."""
        self.clear = options["clear"]
        self.interactive = options["interactive"]

    def handle(self, **options):
        """Collect tools."""
        self.set_options(**options)

        previous_data = any(self.runtime_dir.glob("*"))
        if self.interactive and previous_data:
            self.get_confirmation()

        if self.clear and previous_data:
            self.clear_runtime()

        self.prepare_runtime()



""".. Ignore pydocstyle D400.

==================
Register Processes
==================

"""
import os
import re

import jsonschema
import yaml
from versionfield.utils import convert_version_string_to_int

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.db.models import Max, Q

from resolwe.flow.engine import InvalidEngineError
from resolwe.flow.finders import get_finders
from resolwe.flow.managers import manager
from resolwe.flow.models import DescriptorSchema, Process
from resolwe.flow.models.base import VERSION_NUMBER_BITS
from resolwe.flow.models.utils import validate_schema, validation_schema
from resolwe.flow.utils import dict_dot, iterate_schema
from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions

PROCESSOR_SCHEMA = validation_schema("processor")
DESCRIPTOR_SCHEMA = validation_schema("descriptor")

SCHEMA_TYPE_DESCRIPTOR = "descriptor"
SCHEMA_TYPE_PROCESS = "process"


class Command(BaseCommand):
    """Register processes."""

    help = "Register processes"

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="register also if version mismatch",
        )
        parser.add_argument(
            "--retire",
            default=False,
            action="store_true",
            help="retire obsolete processes",
        )

    def valid(self, instance, schema):
        """Validate schema."""
        try:
            jsonschema.validate(instance, schema)
        except jsonschema.exceptions.ValidationError as ex:
            self.stderr.write(
                "    VALIDATION ERROR: {}".format(
                    instance["name"] if "name" in instance else ""
                )
            )
            self.stderr.write("        path:       {}".format(ex.path))
            self.stderr.write("        message:    {}".format(ex.message))
            self.stderr.write("        validator:  {}".format(ex.validator))
            self.stderr.write("        val. value: {}".format(ex.validator_value))
            return False

        try:
            # Check that default values fit field schema.
            for field in ["input", "output", "schema"]:
                for schema, _, path in iterate_schema({}, instance.get(field, {})):
                    if "default" in schema:
                        validate_schema({schema["name"]: schema["default"]}, [schema])
        except ValidationError:
            self.stderr.write("    VALIDATION ERROR: {}".format(instance["name"]))
            self.stderr.write(
                "        Default value of field '{}' is not valid.".format(path)
            )
            return False

        return True

    def find_descriptor_schemas(self, schema_file):
        """Find descriptor schemas in given path."""
        if not schema_file.lower().endswith((".yml", ".yaml")):
            return []

        with open(schema_file) as fn:
            schemas = yaml.load(fn, Loader=yaml.FullLoader)
        if not schemas:
            self.stderr.write("Could not read YAML file {}".format(schema_file))
            return []

        descriptor_schemas = []
        for schema in schemas:
            if "schema" not in schema:
                continue

            descriptor_schemas.append(schema)

        return descriptor_schemas

    def find_schemas(self, schema_path, schema_type=SCHEMA_TYPE_PROCESS, verbosity=1):
        """Find schemas in packages that match filters."""
        schema_matches = []

        if not os.path.isdir(schema_path):
            if verbosity > 0:
                self.stdout.write("Invalid path {}".format(schema_path))
            return

        if schema_type not in [SCHEMA_TYPE_PROCESS, SCHEMA_TYPE_DESCRIPTOR]:
            raise ValueError("Invalid schema type")

        for root, _, files in os.walk(schema_path):
            for schema_file in [os.path.join(root, fn) for fn in files]:
                schemas = None
                if schema_type == SCHEMA_TYPE_DESCRIPTOR:
                    # Discover descriptors.
                    schemas = self.find_descriptor_schemas(schema_file)
                elif schema_type == SCHEMA_TYPE_PROCESS:
                    # Perform process discovery for all supported execution engines.
                    schemas = []
                    for execution_engine in manager.execution_engines.values():
                        schemas.extend(execution_engine.discover_process(schema_file))

                for schema in schemas:
                    schema_matches.append(schema)

        return schema_matches

    def register_processes(self, process_schemas, user, force=False, verbosity=1):
        """Read and register processors."""
        log_processors = []
        log_templates = []

        for p in process_schemas:
            # TODO: Remove this when all processes are migrated to the
            #       new syntax.
            if "flow_collection" in p:
                if "entity" in p:
                    self.stderr.write(
                        "Skip processor {}: only one of 'flow_collection' and 'entity' fields "
                        "allowed".format(p["slug"])
                    )
                    continue

                p["entity"] = {"type": p.pop("flow_collection")}

            if p["type"][-1] != ":":
                p["type"] += ":"

            if "category" in p and not p["category"].endswith(":"):
                p["category"] += ":"

            for field in ["input", "output"]:
                for schema, _, _ in iterate_schema({}, p[field] if field in p else {}):
                    if not schema["type"][-1].endswith(":"):
                        schema["type"] += ":"
            # TODO: Check if schemas validate with our JSON meta schema and Processor model docs.

            if not self.valid(p, PROCESSOR_SCHEMA):
                continue

            if "entity" in p:
                if "type" not in p["entity"]:
                    self.stderr.write(
                        "Skip process {}: 'entity.type' required if 'entity' defined".format(
                            p["slug"]
                        )
                    )
                    continue
                if "input" in p["entity"] and p["entity"].get("always_create", False):
                    self.stderr.write(
                        "Skip process {}: 'entity.input' will not be considered if 'entity.always_create' "
                        "is set to true.".format(p["slug"])
                    )
                    continue

                p["entity_type"] = p["entity"]["type"]
                p["entity_descriptor_schema"] = p["entity"].get(
                    "descriptor_schema", p["entity_type"]
                )
                p["entity_input"] = p["entity"].get("input", None)
                p["entity_always_create"] = p["entity"].get("always_create", False)
                p.pop("entity")

                if not DescriptorSchema.objects.filter(
                    slug=p["entity_descriptor_schema"]
                ).exists():
                    self.stderr.write(
                        "Skip processor {}: Unknown descriptor schema '{}' used in 'entity' "
                        "field.".format(p["slug"], p["entity_descriptor_schema"])
                    )
                    continue

            if "persistence" in p:
                persistence_mapping = {
                    "RAW": Process.PERSISTENCE_RAW,
                    "CACHED": Process.PERSISTENCE_CACHED,
                    "TEMP": Process.PERSISTENCE_TEMP,
                }

                p["persistence"] = persistence_mapping[p["persistence"]]

            if "scheduling_class" in p:
                scheduling_class_mapping = {
                    "interactive": Process.SCHEDULING_CLASS_INTERACTIVE,
                    "batch": Process.SCHEDULING_CLASS_BATCH,
                }

                p["scheduling_class"] = scheduling_class_mapping[p["scheduling_class"]]

            if "input" in p:
                p["input_schema"] = p.pop("input")

            if "output" in p:
                p["output_schema"] = p.pop("output")

            slug = p["slug"]

            if "run" in p:
                # Set default language to 'bash' if not set.
                p["run"].setdefault("language", "bash")

                # Transform output schema using the execution engine.
                try:
                    execution_engine = manager.get_execution_engine(
                        p["run"]["language"]
                    )
                    extra_output_schema = execution_engine.get_output_schema(p)
                    if extra_output_schema:
                        p.setdefault("output_schema", []).extend(extra_output_schema)
                except InvalidEngineError:
                    self.stderr.write(
                        "Skip processor {}: execution engine '{}' not supported".format(
                            slug, p["run"]["language"]
                        )
                    )
                    continue

            # Validate if container image is allowed based on the configured pattern.
            # NOTE: This validation happens here and is not deferred to executors because the idea
            #       is that this will be moved to a "container" requirement independent of the
            #       executor.
            if hasattr(settings, "FLOW_CONTAINER_VALIDATE_IMAGE"):
                try:
                    container_image = dict_dot(p, "requirements.executor.docker.image")
                    if not re.match(
                        settings.FLOW_CONTAINER_VALIDATE_IMAGE, container_image
                    ):
                        self.stderr.write(
                            "Skip processor {}: container image does not match '{}'".format(
                                slug, settings.FLOW_CONTAINER_VALIDATE_IMAGE
                            )
                        )
                        continue
                except KeyError:
                    pass

            version = p["version"]
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = Process.objects.filter(slug=slug).aggregate(
                Max("version")
            )["version__max"]
            if latest_version is not None and latest_version > int_version:
                self.stderr.write(
                    "Skip processor {}: newer version installed".format(slug)
                )
                continue

            previous_process_qs = Process.objects.filter(slug=slug)
            if previous_process_qs.exists():
                previous_process = previous_process_qs.latest()
            else:
                previous_process = None

            process_query = Process.objects.filter(slug=slug, version=version)
            if process_query.exists():
                if not force:
                    if verbosity > 0:
                        self.stdout.write(
                            "Skip processor {}: same version installed".format(slug)
                        )
                    continue

                process_query.update(**p)
                log_processors.append("Updated {}".format(slug))
            else:
                process = Process.objects.create(contributor=user, **p)
                assign_contributor_permissions(process)
                if previous_process:
                    copy_permissions(previous_process, process)
                log_processors.append("Inserted {}".format(slug))

        if verbosity > 0:
            if log_processors:
                self.stdout.write("Processor Updates:")
                for log in log_processors:
                    self.stdout.write("  {}".format(log))

            if log_templates:
                self.stdout.write("Default Template Updates:")
                for log in log_templates:
                    self.stdout.write("  {}".format(log))

    def register_descriptors(self, descriptor_schemas, user, force=False, verbosity=1):
        """Read and register descriptors."""
        log_descriptors = []

        for descriptor_schema in descriptor_schemas:
            for schema, _, _ in iterate_schema({}, descriptor_schema.get("schema", {})):
                if not schema["type"][-1].endswith(":"):
                    schema["type"] += ":"

            if "schema" not in descriptor_schema:
                descriptor_schema["schema"] = []

            if not self.valid(descriptor_schema, DESCRIPTOR_SCHEMA):
                continue

            slug = descriptor_schema["slug"]
            version = descriptor_schema.get("version", "0.0.0")
            int_version = convert_version_string_to_int(version, VERSION_NUMBER_BITS)

            # `latest version` is returned as `int` so it has to be compared to `int_version`
            latest_version = DescriptorSchema.objects.filter(slug=slug).aggregate(
                Max("version")
            )["version__max"]
            if latest_version is not None and latest_version > int_version:
                self.stderr.write(
                    "Skip descriptor schema {}: newer version installed".format(slug)
                )
                continue

            previous_descriptor_qs = DescriptorSchema.objects.filter(slug=slug)
            if previous_descriptor_qs.exists():
                previous_descriptor = previous_descriptor_qs.latest()
            else:
                previous_descriptor = None

            descriptor_query = DescriptorSchema.objects.filter(
                slug=slug, version=version
            )
            if descriptor_query.exists():
                if not force:
                    if verbosity > 0:
                        self.stdout.write(
                            "Skip descriptor schema {}: same version installed".format(
                                slug
                            )
                        )
                    continue

                descriptor_query.update(**descriptor_schema)
                log_descriptors.append("Updated {}".format(slug))
            else:
                descriptor = DescriptorSchema.objects.create(
                    contributor=user, **descriptor_schema
                )
                assign_contributor_permissions(descriptor)
                if previous_descriptor:
                    copy_permissions(previous_descriptor, descriptor)
                log_descriptors.append("Inserted {}".format(slug))

        if log_descriptors and verbosity > 0:
            self.stdout.write("Descriptor schemas Updates:")
            for log in log_descriptors:
                self.stdout.write("  {}".format(log))

    def retire(self, process_schemas):
        """Retire obsolete processes.

        Remove old process versions without data. Find processes that have been
        registered but do not exist in the code anymore, then:

        - If they do not have data: remove them
        - If they have data: flag them not active (``is_active=False``)

        """
        process_slugs = set(ps["slug"] for ps in process_schemas)

        # Processes that are in DB but not in the code
        retired_processes = Process.objects.filter(~Q(slug__in=process_slugs))

        # Remove retired processes which do not have data
        retired_processes.filter(data__exact=None).delete()

        # Remove non-latest processes which do not have data
        latest_version_processes = Process.objects.order_by(
            "slug", "-version"
        ).distinct("slug")
        Process.objects.filter(data__exact=None).exclude(
            id__in=latest_version_processes
        ).delete()

        # Deactivate retired processes which have data
        retired_processes.update(is_active=False)

    def handle(self, *args, **options):
        """Register processes."""
        force = options.get("force")
        retire = options.get("retire")
        verbosity = int(options.get("verbosity"))

        users = (
            get_user_model().objects.filter(is_superuser=True).order_by("date_joined")
        )

        if not users.exists():
            self.stderr.write("Admin does not exist: create a superuser")
            exit(1)

        process_paths, descriptor_paths = [], []
        process_schemas, descriptor_schemas = [], []

        for finder in get_finders():
            process_paths.extend(finder.find_processes())
            descriptor_paths.extend(finder.find_descriptors())

        for proc_path in process_paths:
            process_schemas.extend(
                self.find_schemas(
                    proc_path, schema_type=SCHEMA_TYPE_PROCESS, verbosity=verbosity
                )
            )

        for desc_path in descriptor_paths:
            descriptor_schemas.extend(
                self.find_schemas(
                    desc_path, schema_type=SCHEMA_TYPE_DESCRIPTOR, verbosity=verbosity
                )
            )

        user_admin = users.first()
        self.register_descriptors(
            descriptor_schemas, user_admin, force, verbosity=verbosity
        )
        # NOTE: Descriptor schemas must be registered first, so
        #       processes can validate 'entity_descriptor_schema' field.
        self.register_processes(process_schemas, user_admin, force, verbosity=verbosity)

        if retire:
            self.retire(process_schemas)

        if verbosity > 0:
            self.stdout.write("Running executor post-registration hook...")
        manager.get_executor().post_register_hook(verbosity=verbosity)



""".. Ignore pydocstyle D400.

========
Consumer
========

Manager Channels consumer.

"""
import asyncio
import logging

from channels.consumer import AsyncConsumer
from channels.layers import get_channel_layer
from channels.testing import ApplicationCommunicator

from . import state

logger = logging.getLogger(__name__)


async def send_event(message):
    """Construct a Channels event packet with the given message.

    :param message: The message to send to the manager workers.
    """
    packet = {
        "type": "control_event",  # This is used as the method name in the consumer.
        "content": message,
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)


async def run_consumer(timeout=None):
    """Run the consumer until it finishes processing.

    :param timeout: Set maximum execution time before cancellation, or
        ``None`` (default) for unlimited.
    """
    manager_channel = state.MANAGER_CONTROL_CHANNEL
    manager_scope = {
        "type": "control_event",
        "channel": manager_channel,
    }
    manager_app = ApplicationCommunicator(ManagerConsumer(), manager_scope)

    listener_channel = state.LISTENER_CONTROL_CHANNEL
    listener_scope = {
        "channel": listener_channel,
    }
    listener_app = ApplicationCommunicator(ListenerConsumer(), listener_scope)

    channel_layer = get_channel_layer()

    async def _consume_loop(channel, scope, app):
        """Run a loop to consume messages off the channels layer."""
        message = await channel_layer.receive(channel)
        while message.get("type", {}) != "_resolwe_manager_quit":
            message.update(scope)
            await app.send_input(message)
            message = await channel_layer.receive(channel)

    awaitables = [
        asyncio.ensure_future(
            _consume_loop(manager_channel, manager_scope, manager_app)
        ),
        asyncio.ensure_future(
            _consume_loop(listener_channel, listener_scope, listener_app)
        ),
    ]

    await asyncio.wait(awaitables, timeout=timeout)
    for awaitable in awaitables:
        awaitable.cancel()

    await manager_app.wait()
    await listener_app.wait()


async def exit_consumer():
    """Cause the synchronous consumer to exit cleanly."""
    packet = {
        "type": "_resolwe_manager_quit",
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)
    await get_channel_layer().send(state.LISTENER_CONTROL_CHANNEL, packet)


class ManagerConsumer(AsyncConsumer):
    """Channels consumer for handling manager events."""

    def __init__(self, *args, **kwargs):
        """Initialize a consumer instance with the given manager."""
        # This import is local in order to avoid startup import loops.
        from . import manager

        self.manager = manager

    async def control_event(self, message):
        """Forward control events to the manager dispatcher."""
        logger.debug("control_event got message %s", message)
        try:
            await self.manager.handle_control_event(message["content"])
        except:
            logger.exception("control_event exception.")


class ListenerConsumer(AsyncConsumer):
    """Channels consumer for handling manager events."""

    def __init__(self, *args, **kwargs):
        """Initialize a consumer instance with the given manager."""
        # This import is local in order to avoid startup import loops.
        from . import listener

        self.listener = listener

    async def terminate_worker_event(self, message):
        """Forward control events to the manager dispatcher."""
        await self.listener.terminate_worker(message["identity"])



"""
Django settings for running tests for Resolwe package.

"""
import os
import sys
from distutils.util import strtobool
from pathlib import Path

PROJECT_ROOT = Path(__file__).parent.resolve()

SECRET_KEY = "secret"

DEBUG = True

MIDDLEWARE_CLASSES = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
)

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "channels",
    # 'kombu.transport.django',  # required for Celery to work with Django DB.
    "rest_framework",
    "guardian",
    "versionfield",
    "resolwe",
    "resolwe.permissions",
    "resolwe.flow",
    "resolwe.storage",
    "resolwe.toolkit",
    "resolwe.test_helpers",
)

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
            ],
        },
    },
]

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    "guardian.backends.ObjectPermissionBackend",
)

ANONYMOUS_USER_NAME = "public"

# Get the current Tox testing environment
# NOTE: This is useful for concurrently running tests with separate environments
toxenv = os.environ.get("TOXENV", "")

# Check if PostgreSQL settings are set via environment variables
pgname = os.environ.get("RESOLWE_POSTGRESQL_NAME", "resolwe")
pguser = os.environ.get("RESOLWE_POSTGRESQL_USER", "resolwe")
pgpass = os.environ.get("RESOLWE_POSTGRESQL_PASS", "resolwe")
pghost = os.environ.get("RESOLWE_POSTGRESQL_HOST", "localhost")
pgport = int(os.environ.get("RESOLWE_POSTGRESQL_PORT", 55432))

DATABASES = {
    "default": {
        "ENGINE": "resolwe.db.postgresql",
        "NAME": pgname,
        "USER": pguser,
        "PASSWORD": pgpass,
        "HOST": pghost,
        "PORT": pgport,
        "TEST": {"NAME": pgname + "_test" + os.environ.get("PYTHONHASHSEED", "")},
    }
}

REDIS_CONNECTION = {
    "host": "localhost",
    "port": int(os.environ.get("RESOLWE_REDIS_PORT", 56379)),
    "db": int(os.environ.get("RESOLWE_REDIS_DATABASE", 1)),
}

LISTENER_CONNECTION = {
    # Keys in the hosts dictionary are workload connector names. Currently
    # supported are 'local', 'kubertenes', 'celery' and 'slurm'.
    "hosts": {"local": "172.17.0.1"},
    "port": int(os.environ.get("RESOLWE_LISTENER_SERVICE_PORT", 53893)),
    "min_port": 50000,
    "max_port": 60000,
    "protocol": "tcp",
}

if sys.platform == "darwin":
    LISTENER_CONNECTION["hosts"]["local"] = "127.0.0.1"

ASGI_APPLICATION = "resolwe.flow.routing.channel_routing"

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [(REDIS_CONNECTION["host"], REDIS_CONNECTION["port"])],
            "expiry": 3600,
        },
    },
}

BROKER_URL = "django://"
CELERY_TASK_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = [CELERY_TASK_SERIALIZER]

STATIC_URL = "/static/"


# Check if any Manager settings are set via environment variables
manager_prefix = os.environ.get("RESOLWE_MANAGER_REDIS_PREFIX", "resolwe.flow.manager")
FLOW_MANAGER = {
    "REDIS_PREFIX": manager_prefix,
    "REDIS_CONNECTION": REDIS_CONNECTION,
    "TEST": {
        "REDIS_PREFIX": manager_prefix + "-test",
    },
}

# Set custom Docker command if set via environment variable.
if "RESOLWE_DOCKER_COMMAND" in os.environ:
    FLOW_DOCKER_COMMAND = os.environ["RESOLWE_DOCKER_COMMAND"]

# Ignore errors when pulling Docker images from 'list_docker_images --pull' command.
FLOW_DOCKER_IGNORE_PULL_ERRORS = strtobool(
    os.environ.get("RESOLWE_DOCKER_IGNORE_PULL_ERRORS", "1")
)

# Don't pull Docker images if set via the environment variable.
FLOW_DOCKER_DONT_PULL = strtobool(os.environ.get("RESOLWE_DOCKER_DONT_PULL", "0"))

# Disable SECCOMP if set via environment variable.
FLOW_DOCKER_DISABLE_SECCOMP = strtobool(
    os.environ.get("RESOLWE_DOCKER_DISABLE_SECCOMP", "0")
)

# Ensure all container images follow a specific format.
FLOW_CONTAINER_VALIDATE_IMAGE = r".+:(?!latest)"

FLOW_API = {
    "PERMISSIONS": "resolwe.permissions.permissions",
}
FLOW_EXPRESSION_ENGINES = [
    {
        "ENGINE": "resolwe.flow.expression_engines.jinja",
        "CUSTOM_FILTERS": [
            "resolwe.flow.tests.expression_filters",
        ],
    },
]
FLOW_EXECUTION_ENGINES = [
    "resolwe.flow.execution_engines.bash",
    "resolwe.flow.execution_engines.python",
    "resolwe.flow.execution_engines.workflow",
]

FLOW_PROCESSES_FINDERS = (
    "resolwe.flow.finders.FileSystemProcessesFinder",
    "resolwe.flow.finders.AppDirectoriesFinder",
)

FLOW_PROCESSES_RUNTIMES = ("resolwe.process.runtime.Process",)

FLOW_EXECUTOR = {
    "NAME": "resolwe.flow.executors.docker",
    "REDIS_CONNECTION": REDIS_CONNECTION,
    "LISTENER_CONNECTION": LISTENER_CONNECTION,
}

FLOW_DOCKER_AUTOREMOVE = False
FLOW_DOCKER_COMMUNICATOR_IMAGE = os.environ.get(
    "RESOLWE_COMMUNICATOR_IMAGE", "public.ecr.aws/s4q6j6e8/resolwe/com:latest"
)
FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE = (
    "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04"
)

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
    ),
    "DEFAULT_FILTER_BACKENDS": (
        "resolwe.permissions.filters.ResolwePermissionsFilter",
        "django_filters.rest_framework.backends.DjangoFilterBackend",
        "resolwe.flow.filters.OrderingFilter",
    ),
    "EXCEPTION_HANDLER": "resolwe.flow.utils.exceptions.resolwe_exception_handler",
    # Python<3.7 cannot parse iso-8601 formatted datetimes with tz-info form
    # "+01:00" (DRF default). It can only parse "+0100" form, so we need to
    # modify this setting. This will be fixed in Python3.7, where "+01:00" can
    # be parsed by ``datetime.datetime.strptime`` syntax.
    # For more, check "%z" syntax description in:
    # https://docs.python.org/3.7/library/datetime.html#strftime-and-strptime-behavior
    "DATETIME_FORMAT": "%Y-%m-%dT%H:%M:%S.%f%z",
}

# Time
USE_TZ = True
TIME_ZONE = "UTC"

# Django does not support parsing of 'iso-8601' formated datetimes by default.
# Since Django-filters uses Django forms for parsing, we need to modify Django
# setting ``DATETIME_INPUT_FORMATS`` to support 'iso-8601' format.
# https://docs.djangoproject.com/en/1.11/ref/settings/#datetime-input-formats
DATETIME_INPUT_FORMATS = (
    # These are already given Django defaults:
    "%Y-%m-%d %H:%M:%S",  # '2006-10-25 14:30:59'
    "%Y-%m-%d %H:%M:%S.%f",  # '2006-10-25 14:30:59.000200'
    "%Y-%m-%d %H:%M",  # '2006-10-25 14:30'
    "%Y-%m-%d",  # '2006-10-25'
    # These are iso-8601 formatted:
    "%Y-%m-%dT%H:%M:%S.%f%z",  # '2006-10-25T14:30:59.000200+0200' or '2006-10-25T14:30:59.000200+02:00' (Python>=3.7)
    "%Y-%m-%dT%H:%M:%S.%fZ",  # '2006-10-25T14:30:59.000200Z'
    "%Y-%m-%dT%H:%M:%S.%f",  # '2006-10-25T14:30:59.000200'
    "%Y-%m-%dT%H:%M:%SZ",  # '2006-10-25T14:30:59Z'
    "%Y-%m-%dT%H:%M:%S",  # '2006-10-25T14:30:59'
    "%Y-%m-%dT%H:%M",  # '2006-10-25T14:30'
)


# Testing.

TEST_RUNNER = "resolwe.test_helpers.test_runner.ResolweRunner"
TEST_PROCESS_REQUIRE_TAGS = True
TEST_PROCESS_PROFILE = False


# Logging.

# Set RESOLWE_LOG_FILE environment variable to a file path to enable
# logging debugging messages to to a file.
debug_file_path = os.environ.get("RESOLWE_LOG_FILE", os.devnull)

github_actions = os.environ.get("GITHUB_ACTIONS") == "true"
CONSOLE_LEVEL = "WARNING"
default_logger_handlers = ["file"]

if github_actions:
    CONSOLE_LEVEL = "DEBUG"
    default_logger_handlers = ["console", "file"]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "%(asctime)s - %(levelname)s - %(name)s[%(process)s]: %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": CONSOLE_LEVEL,
            "formatter": "standard",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": debug_file_path,
            "formatter": "standard",
            "maxBytes": 1024 * 1024 * 10,  # 10 MB
        },
    },
    "loggers": {
        "": {
            "handlers": default_logger_handlers,
            "level": "DEBUG",
        },
    },
}



""".. Ignore pydocstyle D400.

===========
Preparation
===========

.. autoclass:: resolwe.flow.executors.docker.prepare.FlowExecutorPreparer
    :members:

"""
import os

from django.conf import settings
from django.core.management import call_command

from resolwe.storage.connectors import connectors

from .. import constants
from ..prepare import BaseFlowExecutorPreparer


class FlowExecutorPreparer(BaseFlowExecutorPreparer):
    """Specialized manager assist for the docker executor."""

    def post_register_hook(self, verbosity=1):
        """Pull Docker images needed by processes after registering."""
        if not getattr(settings, "FLOW_DOCKER_DONT_PULL", False):
            call_command("list_docker_images", pull=True, verbosity=verbosity)

    def resolve_data_path(self, data=None, filename=None):
        """Resolve data path for use with the executor.

        :param data: Data object instance
        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given data file in programs executed using this executor
        :raises RuntimeError: when data path can not be resolved.
        """
        storage_name = "data"
        filesystem_connectors = [
            connector
            for connector in connectors.for_storage(storage_name)
            if connector.mountable
        ]

        if data is None:
            if not filesystem_connectors:
                return constants.INPUTS_VOLUME
            else:
                return f"/{storage_name}_{filesystem_connectors[0].name}"

        data_connectors = data.location.connectors
        for connector in filesystem_connectors:
            if connector in data_connectors:
                return data.location.get_path(
                    prefix=f"/{storage_name}_{connector.name}", filename=filename
                )

        return data.location.get_path(prefix=constants.INPUTS_VOLUME, filename=filename)

    def resolve_upload_path(self, filename=None):
        """Resolve upload path for use with the executor.

        :param filename: Filename to resolve
        :return: Resolved filename, which can be used to access the
            given uploaded file in programs executed using this
            executor
        """
        upload_connectors = [
            connector
            for connector in connectors.for_storage("upload")
            if connector.mountable
        ]
        if not upload_connectors:
            raise RuntimeError("No connectors are configured for 'upload' storage.")

        upload_connector = upload_connectors[0]
        if filename is None:
            return f"/upload_{upload_connector.name}"

        return f"/upload_{upload_connector.name}/{filename}"

    def get_environment_variables(self):
        """Return dict of environment variables that will be added to executor."""

        return {"TMPDIR": os.fspath(constants.PROCESSING_VOLUME / constants.TMPDIR)}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-10-20 08:12
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.models.utils import hydrate_size


def calculate_size(apps, schema_editor):
    """Compute ``size`` attribute in Data model."""
    Data = apps.get_model("flow", "Data")
    for data in Data.objects.all():
        hydrate_size(data, force=True)
        data.save()


def reverse_migration(apps, schema_editor):
    """
    Reverse migration.

    This function can be emty, since ``size`` field will be deleted in
    next step i.e. previous migration.
    """
    pass


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0007_data_size"),
    ]

    operations = [
        migrations.RunPython(calculate_size, reverse_migration),
    ]



""".. Ignore pydocstyle D400.

====================
Clean test directory
====================

Command to run on local machine::

    ./manage.py cleantestdir

"""
import re
import shutil
from itertools import chain
from pathlib import Path

from django.core.management.base import BaseCommand

from resolwe.storage import settings as storage_settings
from resolwe.storage.connectors import connectors

TEST_DIR_REGEX = r"^test_.*_\d+$"


class Command(BaseCommand):
    """Cleanup files created during testing."""

    help = "Cleanup files created during testing."

    def handle(self, *args, **kwargs):
        """Cleanup files created during testing."""
        directories = [
            Path(connector.path)
            for connector in chain(
                connectors.for_storage("data"), connectors.for_storage("upload")
            )
            if connector.mountable
        ]
        directories += [
            Path(volume_config["config"]["path"])
            for volume_name, volume_config in storage_settings.FLOW_VOLUMES.items()
            if volume_config["config"].get("read_only", False) == False
        ]

        for directory in directories:
            directory = directory.resolve()
            for test_dir in directory.iterdir():
                if not test_dir.is_dir():
                    continue
                if not re.match(TEST_DIR_REGEX, test_dir.name):
                    continue

                shutil.rmtree(test_dir)



""".. Ignore pydocstyle D400.

=====================
Permissions shortcuts
=====================

.. autofunction:: _group_groups
.. autofunction:: get_object_perms

"""
from collections import defaultdict
from itertools import chain, groupby

from django.conf import settings
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Count, Q
from django.shortcuts import _get_queryset

from guardian.compat import get_user_model
from guardian.ctypes import get_content_type
from guardian.exceptions import MixedContentTypeError, WrongAppError
from guardian.models import GroupObjectPermission
from guardian.shortcuts import get_perms, get_users_with_perms
from guardian.utils import (
    get_anonymous_user,
    get_group_obj_perms_model,
    get_identity,
    get_user_obj_perms_model,
)


# XXX: This is a copy of guardian.shortcuts.get_groups_with_perms with a fixed bug
#      Use the original version once the following fix is merged and released:
#      https://github.com/django-guardian/django-guardian/pull/529
def get_groups_with_perms(obj, attach_perms=False):
    """Return queryset of all ``Group`` objects with *any* object permissions for the given ``obj``."""
    ctype = get_content_type(obj)
    group_model = get_group_obj_perms_model(obj)

    if not attach_perms:
        # It's much easier without attached perms so we do it first if that is the case
        group_rel_name = group_model.group.field.related_query_name()
        if group_model.objects.is_generic():
            group_filters = {
                "%s__content_type" % group_rel_name: ctype,
                "%s__object_pk" % group_rel_name: obj.pk,
            }
        else:
            group_filters = {"%s__content_object" % group_rel_name: obj}
        return Group.objects.filter(**group_filters).distinct()
    else:
        group_perms_mapping = defaultdict(list)
        groups_with_perms = get_groups_with_perms(obj)
        queryset = group_model.objects.filter(
            group__in=groups_with_perms
        ).prefetch_related("group", "permission")
        if group_model is GroupObjectPermission:
            queryset = queryset.filter(object_pk=obj.pk, content_type=ctype)
        else:
            queryset = queryset.filter(content_object_id=obj.pk)

        for group_perm in queryset:
            group_perms_mapping[group_perm.group].append(group_perm.permission.codename)
        return dict(group_perms_mapping)


def _group_groups(perm_list):
    """Group permissions by group.

    Input is list of tuples of length 3, where each tuple is in
    following format::

        (<group_id>, <group_name>, <single_permission>)

    Permissions are regrouped and returned in such way that there is
    only one tuple for each group::

        (<group_id>, <group_name>, [<first_permission>, <second_permission>,...])

    :param list perm_list: list of touples of length 3
    :return: list tuples with grouped permissions
    :rtype: list

    """
    perm_list = sorted(perm_list, key=lambda tup: tup[0])

    grouped_perms = []
    for key, group in groupby(perm_list, lambda tup: (tup[0], tup[1])):
        grouped_perms.append((key[0], key[1], [g[2] for g in group]))

    return grouped_perms


def get_user_group_perms(user_or_group, obj):
    """Get permissins for user groups.

    Based on guardian.core.ObjectPermissionChecker.

    """
    user, group = get_identity(user_or_group)

    if user and not user.is_active:
        return [], []
    user_model = get_user_model()
    ctype = ContentType.objects.get_for_model(obj)

    group_model = get_group_obj_perms_model(obj)
    group_rel_name = group_model.permission.field.related_query_name()
    if user:
        user_rel_name = user_model.groups.field.related_query_name()
        group_filters = {user_rel_name: user}
    else:
        group_filters = {"pk": group.pk}
    if group_model.objects.is_generic():
        group_filters.update(
            {
                "{}__content_type".format(group_rel_name): ctype,
                "{}__object_pk".format(group_rel_name): obj.pk,
            }
        )
    else:
        group_filters["{}__content_object".format(group_rel_name)] = obj

    user_perms, group_perms = [], []

    if user:
        perms_qs = Permission.objects.filter(content_type=ctype)
        if user.is_superuser:
            user_perms = list(chain(perms_qs.values_list("codename", flat=True)))
        else:
            model = get_user_obj_perms_model(obj)
            related_name = model.permission.field.related_query_name()
            user_filters = {"{}__user".format(related_name): user}
            if model.objects.is_generic():
                user_filters.update(
                    {
                        "{}__content_type".format(related_name): ctype,
                        "{}__object_pk".format(related_name): obj.pk,
                    }
                )
            else:
                user_filters["{}__content_object".format(related_name)] = obj

            user_perms_qs = perms_qs.filter(**user_filters)
            user_perms = list(chain(user_perms_qs.values_list("codename", flat=True)))

    group_perms_qs = Group.objects.filter(**group_filters)
    group_perms = list(
        chain(
            group_perms_qs.order_by("pk").values_list(
                "pk", "name", "{}__permission__codename".format(group_rel_name)
            )
        )
    )

    group_perms = _group_groups(group_perms)

    return user_perms, group_perms


def get_object_perms(obj, user=None):
    """Return permissions for given object in Resolwe specific format.

    Function returns permissions for given object ``obj`` in following
    format::

       {
           "type": "group"/"user"/"public",
           "id": <group_or_user_id>,
           "name": <group_or_user_name>,
           "permissions": [<first_permission>, <second_permission>,...]
       }

    For ``public`` type ``id`` and ``name`` keys are omitted.

    If ``user`` parameter is given, permissions are limited only to
    given user, groups he belongs to and public permissions.

    :param obj: Resolwe's DB model's instance
    :type obj: a subclass of :class:`~resolwe.flow.models.base.BaseModel`
    :param user: Django user
    :type user: :class:`~django.contrib.auth.models.User` or :data:`None`
    :return: list of permissions object in described format
    :rtype: list

    """

    def format_permissions(perms):
        """Remove model name from permission."""
        ctype = ContentType.objects.get_for_model(obj)
        return [perm.replace("_{}".format(ctype.name), "") for perm in perms]

    perms_list = []

    if user:
        if user.is_authenticated:
            user_perms, group_perms = get_user_group_perms(user, obj)
        else:
            user_perms, group_perms = [], []

        if user_perms != []:
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": format_permissions(user_perms),
                }
            )

        if group_perms != []:
            for group_id, group_name, perms in group_perms:
                perms_list.append(
                    {
                        "type": "group",
                        "id": group_id,
                        "name": group_name,
                        "permissions": format_permissions(perms),
                    }
                )
    else:
        user_options = {"attach_perms": True, "with_group_users": False}
        for user, perms in get_users_with_perms(obj, **user_options).items():
            if user.username == settings.ANONYMOUS_USER_NAME:
                # public user is treated separately
                continue
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": format_permissions(perms),
                }
            )

        group_options = {
            "attach_perms": True,
        }
        for group, perms in get_groups_with_perms(obj, **group_options).items():
            perms_list.append(
                {
                    "type": "group",
                    "id": group.pk,
                    "name": group.name,
                    "permissions": format_permissions(perms),
                }
            )

    public_perms = get_perms(AnonymousUser(), obj)
    if public_perms != []:
        perms_list.append(
            {
                "type": "public",
                "permissions": format_permissions(public_perms),
            }
        )

    return perms_list


# based on guardian.shortcuts.get_objects_for_user
def get_objects_for_user(
    user,
    perms,
    klass=None,
    use_groups=True,
    any_perm=False,
    with_superuser=True,
    accept_global_perms=True,
    perms_filter="pk__in",
):
    """Return queryset with required permissions."""
    if isinstance(perms, str):
        perms = [perms]

    ctype = None
    app_label = None
    codenames = set()

    # Compute codenames set and ctype if possible
    for perm in perms:
        if "." in perm:
            new_app_label, codename = perm.split(".", 1)
            if app_label is not None and app_label != new_app_label:
                raise MixedContentTypeError(
                    "Given perms must have same app label "
                    "({} != {})".format(app_label, new_app_label)
                )
            else:
                app_label = new_app_label
        else:
            codename = perm
        codenames.add(codename)

        if app_label is not None:
            new_ctype = ContentType.objects.get(
                app_label=app_label, permission__codename=codename
            )
            if ctype is not None and ctype != new_ctype:
                raise MixedContentTypeError(
                    "ContentType was once computed to be {} and another "
                    "one {}".format(ctype, new_ctype)
                )
            else:
                ctype = new_ctype

    # Compute queryset and ctype if still missing
    if ctype is None and klass is not None:
        queryset = _get_queryset(klass)
        ctype = ContentType.objects.get_for_model(queryset.model)
    elif ctype is not None and klass is None:
        queryset = _get_queryset(ctype.model_class())
    elif klass is None:
        raise WrongAppError("Cannot determine content type")
    else:
        queryset = _get_queryset(klass)
        if ctype.model_class() != queryset.model and perms_filter == "pk__in":
            raise MixedContentTypeError(
                "Content type for given perms and " "klass differs"
            )

    # At this point, we should have both ctype and queryset and they should
    # match which means: ctype.model_class() == queryset.model
    # we should also have `codenames` list

    # First check if user is superuser and if so, return queryset immediately
    if with_superuser and user.is_superuser:
        return queryset

    # Check if the user is anonymous. The
    # django.contrib.auth.models.AnonymousUser object doesn't work for queries
    # and it's nice to be able to pass in request.user blindly.
    if user.is_anonymous:
        user = get_anonymous_user()

    global_perms = set()
    has_global_perms = False
    # a superuser has by default assigned global perms for any
    if accept_global_perms and with_superuser:
        for code in codenames:
            if user.has_perm(ctype.app_label + "." + code):
                global_perms.add(code)
        for code in global_perms:
            codenames.remove(code)
        # prerequisite: there must be elements in global_perms otherwise just
        # follow the procedure for object based permissions only AND
        # 1. codenames is empty, which means that permissions are ONLY set
        # globally, therefore return the full queryset.
        # OR
        # 2. any_perm is True, then the global permission beats the object
        # based permission anyway, therefore return full queryset
        if global_perms and (not codenames or any_perm):
            return queryset
        # if we have global perms and still some object based perms differing
        # from global perms and any_perm is set to false, then we have to flag
        # that global perms exist in order to merge object based permissions by
        # user and by group correctly. Scenario: global perm change_xx and
        # object based perm delete_xx on object A for user, and object based
        # permission delete_xx  on object B for group, to which user is
        # assigned.
        # get_objects_for_user(user, [change_xx, delete_xx], use_groups=True,
        # any_perm=False, accept_global_perms=True) must retrieve object A and
        # B.
        elif global_perms and codenames:
            has_global_perms = True

    # Now we should extract list of pk values for which we would filter
    # queryset
    user_model = get_user_obj_perms_model(queryset.model)
    user_obj_perms_queryset = user_model.objects.filter(
        Q(user=user) | Q(user=get_anonymous_user())
    ).filter(permission__content_type=ctype)

    if codenames:
        user_obj_perms_queryset = user_obj_perms_queryset.filter(
            permission__codename__in=codenames
        )
    direct_fields = ["content_object__pk", "permission__codename"]
    generic_fields = ["object_pk", "permission__codename"]
    if user_model.objects.is_generic():
        user_fields = generic_fields
    else:
        user_fields = direct_fields

    if use_groups:
        group_model = get_group_obj_perms_model(queryset.model)
        group_filters = {
            "permission__content_type": ctype,
            "group__{}".format(
                get_user_model().groups.field.related_query_name()
            ): user,
        }
        if codenames:
            group_filters.update(
                {
                    "permission__codename__in": codenames,
                }
            )
        groups_obj_perms_queryset = group_model.objects.filter(**group_filters)
        if group_model.objects.is_generic():
            group_fields = generic_fields
        else:
            group_fields = direct_fields
        if not any_perm and codenames and not has_global_perms:
            user_obj_perms = user_obj_perms_queryset.values_list(*user_fields)
            groups_obj_perms = groups_obj_perms_queryset.values_list(*group_fields)
            data = list(user_obj_perms) + list(groups_obj_perms)
            # sorting/grouping by pk (first in result tuple)
            data = sorted(data, key=lambda t: t[0])
            pk_list = []
            for pk, group in groupby(data, lambda t: t[0]):
                obj_codenames = set((e[1] for e in group))
                if codenames.issubset(obj_codenames):
                    pk_list.append(pk)
            objects = queryset.filter(**{perms_filter: pk_list})
            return objects

    if not any_perm and len(codenames) > 1:
        counts = user_obj_perms_queryset.values(user_fields[0]).annotate(
            object_pk_count=Count(user_fields[0])
        )
        user_obj_perms_queryset = counts.filter(object_pk_count__gte=len(codenames))

    values = user_obj_perms_queryset.values_list(user_fields[0], flat=True)
    if user_model.objects.is_generic():
        values = list(values)
    query = Q(**{perms_filter: values})
    if use_groups:
        values = groups_obj_perms_queryset.values_list(group_fields[0], flat=True)
        if group_model.objects.is_generic():
            values = list(values)
        query |= Q(**{perms_filter: values})

    return queryset.filter(query)


def get_users_with_permission(obj, permission):
    """Return users with specific permission on object.

    :param obj: Object to return users for
    :param permission: Permission codename
    """
    user_model = get_user_model()
    return user_model.objects.filter(
        userobjectpermission__object_pk=obj.pk,
        userobjectpermission__permission__codename=permission,
    ).distinct()



# Generated by Django 2.2.10 on 2021-02-02 09:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("storage", "0008_accesslog_cause"),
    ]

    operations = [
        migrations.AddField(
            model_name="referencedpath",
            name="chunk_size",
            field=models.IntegerField(default=8388608),
        ),
    ]



"""Resolwe models duplicate utils."""
from copy import deepcopy

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import IntegrityError, models, transaction
from django.utils import timezone

from guardian.models import GroupObjectPermission, UserObjectPermission

from resolwe.flow.utils import iterate_fields


def _check_permissions(objects, permission, user):
    """Check that user has permission on all given objects.

    :param objects: A list of objects to be checked.
    :type objects: list of `~django.db.models.Model`

    :param str permission: Codename of the required permission.

    :param contributor: A Django user that should have the permission.
    :type contributor: `~django.contrib.auth.models.User`

    :raises: ValidationError when user does not have all the required
        permissions.
    """
    for obj in objects:
        if not user.has_perm(permission, obj):
            raise ValidationError(
                "User doesn't have '{}' permission on {}.".format(permission, obj)
            )


def _bulk_copy_permissions(content_type, pk_mapping):
    """Copy permissions from old objects to the new ones.

    Given primary key mapping is a mapping between old object ids, from which
    permissions are copied, and new object ids, to which permissions are
    copied.

    :param content_type: Content type of objects to process.
    :type content_type: `~django.contrib.auth.models.ContentType`

    :param dict pk_mapping: Dictionary with old primary keys as keys and new
        primary keys as values.

    """
    for permission_cls in [GroupObjectPermission, UserObjectPermission]:

        permissions_qs = permission_cls.objects.filter(
            content_type=content_type, object_pk__in=pk_mapping.keys()
        )
        new_permissions = []

        for permission in permissions_qs:
            permission.pk = None
            permission.object_pk = str(pk_mapping[int(permission.object_pk)])
            new_permissions.append(permission)
        permission_cls.objects.bulk_create(new_permissions)


def _bulk_assign_contributor_permission(content_type, pks, user):
    """Assign all available permissions on given objects to given user.

    Objects are determined by the combination of the content type and a
    list of primary keys.

    :param content_type: Content type of objects to process.
    :type content_type: `~django.contrib.auth.models.ContentType`

    :param list pks: List of primary keys of objects to process.

    :param user: A Django user to which permissions should be assigned.
    :type contributor: `~django.contrib.auth.models.User`

    """
    permissions = Permission.objects.filter(content_type=content_type)

    new_permissions = []
    for permission in permissions:
        for pk in pks:
            new_permissions.append(
                UserObjectPermission(
                    content_type=content_type,
                    object_pk=pk,
                    permission=permission,
                    user=user,
                )
            )
    UserObjectPermission.objects.bulk_create(new_permissions, ignore_conflicts=True)


def _rewire_foreign_key(objects, target, pk_mapping):
    """Make a copy of given objects and rewire references to target objects.

    All fields that points to an object of ``target`` type are replaced with
    the new objects according to the provided mapping. The list of new objects
    is returned.

    :param objects: A list of objects to process.
    :type objects: list of `~django.db.models.Model`

    :param target: A class to which references should be replaced
    :type terget: `~django.db.models.Model`

    :param dict pk_mapping: Dictionary with old primary keys as keys and new
        primary keys as values.

    :rtype: list of `~django.db.models.Model`

    """
    if not objects:
        return []

    # Check that all objects are instances of the same class.
    obj_class = objects[0].__class__
    if any(obj.__class__ != obj_class for obj in objects):
        raise ValueError("All objects must be instances of the same class.")

    # Find fields pointing to the target model.
    to_map = [
        field.name
        for field in obj_class._meta.fields
        if type(field) == models.ForeignKey and field.related_model == target
    ]

    # Rewire references.
    new_objects = []
    for obj in objects:
        new_obj = deepcopy(obj)
        for field_name in to_map:
            field_name = "{}_id".format(field_name)
            value = getattr(obj, field_name)
            if value and value in pk_mapping:
                setattr(new_obj, field_name, pk_mapping[value])

        new_obj.pk = None
        new_objects.append(new_obj)

    return obj_class.objects.bulk_create(new_objects)


def _rewire_inputs(data, pk_mapping):
    """Rewire inputs of provided data objects.

    References to input data objects in ``input`` field of the Data objects are
    chenged with references of their copies according to the provided mapping.
    If primary key is not in the mapping, original value is kept.

    :param list data: A list of Data objects to process.

    :param dict pk_mapping: A dict where keys are primary keys of original Data
        objects and values are primary keys of their copies.

    """
    for datum in data:
        for field_schema, fields in iterate_fields(
            datum.input, datum.process.input_schema
        ):
            name = field_schema["name"]
            value = fields[name]

            if field_schema["type"].startswith("data:") and value in pk_mapping:
                fields[name] = pk_mapping[value]

            elif field_schema["type"].startswith("list:data:"):
                fields[name] = [
                    pk_mapping[pk] if pk in pk_mapping else pk for pk in value
                ]


def process_entity(inherit_collection, collection_mapping):
    """Return a function that process a duplicated Entity before saving.

    :param bool inherit_collection: Indicates whether the collection should
        be inherited or not.

    :param dict collection_mapping: A dict where keys are primary keys of
        original collections and values are primary keys of their copies.

    """

    def processor(entity):
        """Set collection of given Entity.

        Collection is to ``None`` if ``inherit_collection`` value is set to
        ``False``, or mapped according to ``collection_mapping`` otherwise (or
        left as is if value is not in the mapping).

        :param entity: An entity to be processed.
        :type entity: `~resolwe.flow.models.Entity`

        :rtype: `~resolwe.flow.models.Entity`

        """
        if inherit_collection and entity.collection:
            if entity.collection.id in collection_mapping:
                entity.collection = collection_mapping[entity.collection.pk]
        else:
            entity.collection = None

        return entity

    return processor


def process_data(
    inherit_collection, inherit_entity, collection_mapping, entity_mapping
):
    """Return a function that process a duplicated Data object before saving.

    :param bool inherit_collection: Indicates whether the collection should
        be inherited or not.

    :param bool inherit_entity: Indicates whether the entity should be
        inherited or not.

    :param dict collection_mapping: A dict where keys are primary keys of
        original collections and values are primary keys of their copies.

    :param dict entity_mapping: A dict where keys are primary keys of original
        entities and values are primary keys of their copies.

    """

    def processor(datum):
        """Set collection and entity of given Data object.

        Collection and entity are set to ``None`` if ``inherit_collection``
        and ``inherit_entity`` values are set to ``False``, respectively, and
        mapped according to ``collection_mapping`` and ``entity_mapping``
        otherwise (or left as is if value is not in the mapping).

        :param datum: An Dat object to be processed.
        :type datum: `~resolwe.flow.models.Data`

        :rtype: `~resolwe.flow.models.Data`

        """
        if inherit_collection and datum.collection:
            if datum.collection.id in collection_mapping:
                datum.collection = collection_mapping[datum.collection.pk]
        else:
            datum.collection = None

        if inherit_entity and datum.entity:
            if datum.entity.id in entity_mapping:
                datum.entity = entity_mapping[datum.entity.pk]
        else:
            datum.entity = None

        return datum

    return processor


def copy_objects(objects, contributor, name_prefix, obj_processor=None):
    """Make a copy of given queryset.

    Shallow copy given queryset and set contributor to the given value,
    prepend name with the prefix, set slug to a unique value, and set
    ``duplicated`` date to the current time. Special attention is paid
    to keep the ``created`` date to its original value.

    If ``obj_processor`` function is given, each object is passed to it
    and the return value is used instead of it.

    :param objects: A queryset to be copied.
    :type objects: `~resolwe.flow.models.base.BaseQuerySet`

    :param contributor: A Django user that will be assigned to copied objects
        as contributor.
    :type contributor: `~django.contrib.auth.models.User`

    :param str name_prefix: A prefix that will be prepend to the name of copied
        objects.

    """
    first = objects.first()
    if not first:
        return objects

    name_max_length = first._meta.get_field("name").max_length
    content_type = ContentType.objects.get_for_model(first)
    model = first._meta.model

    new_objects = []
    for obj in objects:
        new_obj = deepcopy(obj)
        new_obj.pk = None
        new_obj.slug = None
        new_obj.contributor = contributor
        new_obj.name = "{} {}".format(name_prefix, obj.name)

        if len(new_obj.name) > name_max_length:
            new_obj.name = "{}...".format(new_obj.name[: name_max_length - 3])

        if obj_processor:
            new_obj = obj_processor(new_obj)

        new_objects.append(new_obj)

    try:
        # Add another atomic block to avoid corupting the main one.
        with transaction.atomic():
            model.objects.bulk_create(new_objects)
    except IntegrityError:
        # Probably a slug collision occured, try to create objects one by one.
        for obj in new_objects:
            obj.slug = None
            # Call the parent method to skip pre-processing and validation.
            models.Model.save(obj)

    for old, new in zip(objects, new_objects):
        new.created = old.created
        new.duplicated = timezone.now()

    model.objects.bulk_update(new_objects, ["created", "duplicated"])

    pk_mapping = {old.pk: new.pk for old, new in zip(objects, new_objects)}
    _bulk_copy_permissions(content_type, pk_mapping)
    _bulk_assign_contributor_permission(content_type, pk_mapping.values(), contributor)

    return new_objects


@transaction.atomic
def bulk_duplicate(
    collections=None,
    entities=None,
    data=None,
    contributor=None,
    inherit_collection=False,
    inherit_entity=False,
    name_prefix=None,
):
    """Make a copy of given collection, entity or data queryset.

    Exactly one of ``collections``, ``entities`` and ``data`` parameters should
    be passed to the function and should respectively represent a queryset of
    Collection, Entity and Data objects to be copied.

    When copying Collections or Entities, also the contained objects (Entities
    and Data) are also copied.

    Copied objects are transformed in the following ways:

    * ``name_prefix`` ("Copy of " by default) string is prepend to names of all
      copied objects
    * Collection and/or entity of top-most copied objects are preserved only if
      ``inherit_collection`` and/or ``inherit_entity`` values are set to
      ``True``
    * all contained objects are attached to new collections and entities
    * ``input`` fields of all copied Data objects are processed and all inputs
      are replaced with their copies if they exist
    * permissions are copied from original objects
    * Data migration history is copied and linked to the new Data objects

    :param collections: A collection queryset to duplicate.
    :type collections: `~resolwe.flow.models.collection.CollectionQuerySet`

    :param entities: An entity queryset to duplicate.
    :type entity: `~resolwe.flow.models.entity.EntityQuerySet`

    :param data: A data queryset to duplicate.
    :type data: `~resolwe.flow.models.data.DataQuerySet`

    :param contributor: A Django user that will be assigned to copied objects
        as contributor.
    :type contributor: `~django.contrib.auth.models.User`

    :param bool inherit_collection: Indicates whether copied entities and data
        objects are added to the same collection as originals or not.

    :param bool inherit_entity: Indicates whether copied data  objects are
        added to the same collection as originals or not.

    :param str name_prefix: A prefix that will be prepend to the name of all
        copied objects.

    :rtype: `~resolwe.flow.models.collection.CollectionQuerySet` or
        `~resolwe.flow.models.entity.EntityQuerySet` or
        `~resolwe.flow.models.data.DataQuerySet`

    """
    # Prevent circular import.
    from resolwe.flow.models import (
        Collection,
        Data,
        DataDependency,
        DataMigrationHistory,
        Entity,
    )

    name_prefix = name_prefix or "Copy of"

    if not contributor:
        raise ValueError("Contributor must be given.")

    specified = [model for model in [collections, entities, data] if model is not None]
    if len(specified) != 1:
        raise ValueError(
            "Exactly one of 'collections', 'entites', and 'data' attributes must be given."
        )
    if not isinstance(specified[0], models.QuerySet):
        raise ValueError("Function only supports duplcating querysets.")

    collection_mapping, entity_mapping = {}, {}

    if collections is not None:
        new_collections = copy_objects(collections, contributor, name_prefix)

        collection_mapping = {
            old.pk: new for old, new in zip(collections, new_collections)
        }

        entities = Entity.objects.filter(collection__in=collections)
        data = Data.objects.filter(collection__in=collections)

        inherit_collection = True

    if entities is not None:
        entity_processor = process_entity(inherit_collection, collection_mapping)
        new_entities = copy_objects(
            entities, contributor, name_prefix, entity_processor
        )

        entity_mapping = {old.pk: new for old, new in zip(entities, new_entities)}

        # Entity data is also included in the collection, so we don't need to set it
        # here if it was already set (i.e. if we are duplicating a collection).
        if data is None:
            data = Data.objects.filter(entity__in=entities)

        inherit_entity = True

    if data.exclude(status__in=[Data.STATUS_DONE, Data.STATUS_ERROR]).exists():
        raise ValidationError(
            "Data object must have done or error status to be duplicated."
        )

    data_processor = process_data(
        inherit_collection, inherit_entity, collection_mapping, entity_mapping
    )
    new_data = copy_objects(data, contributor, name_prefix, data_processor)

    data_pks = data.values_list("pk")
    pk_mapping = {old.pk: new.pk for old, new in zip(data, new_data)}

    _rewire_inputs(new_data, pk_mapping)
    Data.objects.bulk_update(new_data, ["input"])

    # Add m2m storage relations.
    through_model = Data.storages.through
    storage_relations = through_model.objects.filter(data__in=data_pks)
    _rewire_foreign_key(storage_relations, Data, pk_mapping)

    # Copy data dependencies.
    # We only inherit parent dependencies to keep track of the Data object's
    # source. Child dependencies reflect later actions that don't directly
    # depend on the current object, so copying them would only make unnecessary
    # noise. Duplicate dependencies are not copied to keep a simple chain with
    # links only to direct predecessors.
    data_dependencies = DataDependency.objects.filter(
        kind__in=[DataDependency.KIND_IO, DataDependency.KIND_SUBPROCESS],
        child__in=data_pks,
    )
    _rewire_foreign_key(data_dependencies, Data, pk_mapping)

    # Add duplicate dependencies.
    DataDependency.objects.bulk_create(
        DataDependency(parent=old, child=new, kind=DataDependency.KIND_DUPLICATE)
        for old, new in zip(data, new_data)
    )

    # Copy migration history. Ordering is needed to keep the order of migrations.
    migration_history = DataMigrationHistory.objects.filter(data__in=data_pks).order_by(
        "pk"
    )
    _rewire_foreign_key(migration_history, Data, pk_mapping)

    if inherit_collection:
        data_collections = Collection.objects.filter(data__in=new_data)
        _check_permissions(data_collections, "edit_collection", contributor)
        if entities is not None:
            entity_collections = Collection.objects.filter(entity__in=new_entities)
            _check_permissions(entity_collections, "edit_collection", contributor)

    if inherit_entity:
        data_entities = Entity.objects.filter(data__in=new_data)
        _check_permissions(data_entities, "edit_entity", contributor)

    if collections is not None:
        return new_collections
    elif entities is not None:
        return new_entities
    return new_data



# pylint: disable=missing-docstring
from django.urls import reverse

from rest_framework.test import APIRequestFactory

from resolwe.test import TestCase


class CollectionResourceTest(TestCase):
    def setUp(self):
        super().setUp()

        self.api_client = APIRequestFactory()
        self.list_url = reverse("resolwe-api:collection-list")
        # TODO: add object's id
        # self.detail_url = reverse('resolwe-api:collection-detail')

    def test_readonly_fields(self):
        self.api_client.get(self.list_url)



"""Commands specific to kubernetes installation."""

from typing import TYPE_CHECKING

from resolwe.flow.executors.socket_utils import Message, Response

from .plugin import ListenerPlugin

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor


class InitContainerPlugin(ListenerPlugin):
    """Handler methods for Init Container."""

    name = "Init container plugin"

    def handle_init_suspend_heartbeat(
        self, message: Message[int], manager: "Processor"
    ) -> Response[str]:
        """Suspend heartbeat."""
        manager._listener.communicator.suspend_heartbeat(manager.peer_identity)
        return message.respond_ok("")
